!
! alignmol --- determine alignment of molecules by a laser pulse
! Christian Buth, Robin Santra
! 21 Mar 2007
!
! Version 1.0.0: 21 Mar 2007 (Taken over symmetric rotor code from
!       Robin Santra)
! Version 1.1.0: 03 Apr 2007 (Implemented ratio of x-ray
!       photoabsorption cross section for parallel versus perpendicular
!       polarizations with respect to the laser polarization axis)
! Version 1.2.0: 11 Apr 2007 (Specialization to the linear rotor)
! Version 1.3.0: 13 Jul 2007 (Removed bug from direction cosine
!       matrix elements: the wrong handling of zero angular momenta.
!       Run modes introduced.  A non zero ratio for parallel /
!       perpendicular dipole component)
! Version 1.4.0: 19 Jul 2007 (Removed bug from r_thermal computation;
!       the thermal density matrix was not calculated starting with J=0)
!


      program alignmol
      use mathtl
      use physconst
      use ridgidrotor
      use dirCosRotMatEl
      implicit none

      ! ncols is the number of J,K,M states (J=0,...,Jmax; K = -J,...,J; M = -J,...,J).
      integer ncols
      ! ntime is the number of time steps used for integrating
      ! the rate equations
      integer ntime
      ! Number of x-ray pulses placed along the propagation time span
      integer nxpulses
      ! What parts of the code shall be executed
      integer runmode

      integer timestep, timedata, plotpoints, i

      real(dblprec) upper_R_pp, upper_R_par_thermal, upper_R_perp_thermal,  &
     &          intR_par_perp, intR_par_thermal, intR_perp_thermal
      real(dblprec) Delta, temperature, partitionFunc, theta, thetarad, ratio,  &
     &          time_beg, time_end, t0, tf, dt, Delz2, dipRatio
      real(dblprec), allocatable :: xpcenter(:)

      complex(dblprec) intR_par, intR_perp, intR_thermal, intR_par_curr,  &
     &          intR_perp_curr, intR_thermal_curr, intR_par_prev,         &
     &          intR_perp_prev, intR_thermal_prev
      complex(dblprec) cos2t
      complex(dblprec), allocatable ::                                 &
     &          ai_par_curr(:), ai_perp_curr(:), ai_thermal_curr(:),   &
     &          ai_par(:), ai_perp(:), ai_thermal(:),                  &
     &          ai_par_prev(:), ai_perp_prev(:), ai_thermal_prev(:),   &
     &          ai_par_perp(:), ai_par_thermal(:), ai_perp_thermal(:)

      external REQ

      namelist/AlignMolecule/ RotorType, nuclstat, Jmax, ntime, neq, PulseEnergy,   &
     &          Nxrays, xrayDelta, tau_xr, temperature, RotConstA, RotConstB,       &
     &          PolAni, time_beg, time_end, tau, Delta, laserpulseshape, nxpulses,  &
     &          runmode, dipRatio, seqInt, numPulse, plotpoints, homonuclspin

      ! Initialization
      write (*,*) 'alignmol V1.4.0, alignment of molecules by an intense laser'
      write (*,*) '(c) 2007 Christian Buth, Robin Santra'
      write (*,*)

      ! Default values for parameters
      runmode         = 2
      RotorType       = 2
      nuclstat        = 3
      Jmax            = 13
      ntime           = 2000
      neq             = 20
      laserpulseshape = 1
      PulseEnergy     = 2.0D0
      tau             = 110.0D0
      Delta           = 60.0D0
      Nxrays          = 2.5D5
      xrayDelta       = 10.0D0
      nxpulses        = 300
      tau_xr          = 100.0D0
      temperature     = 10.0D0
      RotConstA       = 0.191205D0
      RotConstB       = 0.069834D0
      PolAni          = 12.0D0
      time_beg        = -400.0D0
      time_end        = 400.0D0
      dipRatio        = 0.0D0
      numPulse        = 50
      seqInt          = 2.0D0
      plotpoints      = -1
      homonuclspin    = 1.0D0

      read(*,nml=AlignMolecule)

      select case (runmode)
        case(1)
          write (*,*) 'Perform following calculation: cos^2 theta'
        case(2)
          write (*,*) 'Perform following calculations: cross section, cos^2 theta'
        case(3)
          write (*,*) 'Determine thermal partition function only'
        case default
          stop 'alignmol: Unkown run mode requested!'
      end select

      write (*,*) 'Maximum angular momentum in the initial density matrix', Jmax
      ! Make sure, t=0 is a data point in the time evolution
      i = int((0.0D0 - time_beg) / (time_end - time_beg) * real(ntime, dblprec)) + 1
      ntime = int((time_end - time_beg) / (0.0D0 - time_beg) * real(i, dblprec))
      write (*,*) 'Number of time steps (adjusted such that t = 0 is a grid point)', ntime
      write (*,*) 'Number of angular momenta in density matrix propagation', neq
      write (*,*) 'Energy of laser pulse', PulseEnergy, 'mJ'
      write (*,*) 'Laser pulse evelope duration (FWHM)', tau, 'ps'
      write (*,*) 'Laser pulse evelope diameter (FWHM)', Delta, 'mum'

      select case (laserpulseshape)
        case(1)
          write (*,*) 'Gaussian laser pulse'
        case(2)
          write (*,*) 'Gaussian laser pulse with two superimposed oscillations'
        case(3)
          write (*,*) 'Laser pulse with sech^2 form'
        case(4)
          write (*,*) 'Sequence of Gaussian laser pulses'
          write (*,*) 'Number of pulses', numPulse
          write (*,*) 'Spacing between pulses ', seqInt, ' ps'
          seqInt = seqInt * PicosecondTot0
        case(5)
          write (*,*) 'Gaussian laser pulse with wiggles on top'
        case default
          stop 'alignmol: Unkown laser pulse shape requested!'
      end select

      if (runmode == 2) then
        write (*,*) 'Number of x-ray photons per pulse', Nxrays
        write (*,*) 'Gaussian x-ray pulse diameter (FWHM)', xrayDelta, 'mum'
        write (*,*) 'Gaussian x-ray pulse duration (FWHM)', tau_xr, 'ps'
        write (*,*) 'Parallel/perpendicular dipole component squared', dipRatio
        ! Make sure, t=0 is a data point in the cross correlations
        i = int((0.0D0 - time_beg) / (time_end - time_beg) * real(nxpulses, dblprec)) + 1
        nxpulses = int((time_end - time_beg) / (0.0D0 - time_beg) * real(i, dblprec)) + 1
        write (*,*) 'Number of Gaussians used to compute the cross correlations ', nxpulses
      end if

      write (*,*) 'Rotational temperature of molecules', temperature, 'K'
      select case (RotorType)
        case(1)
          if (nuclstat /= 1 .and. nuclstat /= 2) then
            stop 'alignmol: Nuclear statistics is incompatible with a linear molecule!'
          end if
          if (RotConstA /= 0.0D0) then
            stop 'alignmol: Linear molecules do not have a RotConstA /= 0.0D0!'
          end if
          write (*,*) 'Molecule is linear'
          write (*,*) 'Rotational constant B', RotConstB, 'cm^-1'
          Kmax = 0              ! There is no k quantum number for linear molecules
        case(2)
          if (nuclstat /= 3) then
            stop 'alignmol: Nuclear statistics is incompatible with a symmetric top molecule!'
          end if
          write (*,*) 'Molecule is a prolate symmetric top'
          write (*,*) 'Rotational constant A', RotConstA, 'cm^-1'
          write (*,*) 'Rotational constant B', RotConstB, 'cm^-1'
          Kmax = Jmax
        case default
          stop 'alignmol: Undefined rotor type!'
      end select

      select case (nuclstat)
        case(1)
          write (*,*) 'Linear molecule with nuclear statistics for',  &
     &          'isotopically identical nuclei with nuclear spin ', homonuclspin
        case(2)
          write (*,*) 'Linear molecule has nuclear statistics of two distinct nuclei'
        case(3)
          write (*,*) 'Symmetric top molecule has CF3Br/CH3I nuclear statistics'
        case default
          stop 'alignmol: Undefined nuclear statistics type!'
      end select

      write (*,*) 'Polarizability anisotropy', PolAni, 'Bohr^3'
      write (*,*) 'Start time for propagation', time_beg, 'ps'
      write (*,*) 'End time for propagation', time_end, 'ps'

      ! Number of points to plot
      if (plotpoints == -1) then
        plotpoints = ntime
      else if (plotpoints > ntime) then
        write (*,*) 'Only one point per time step can be producted for plotting!'
        stop 'alignmol: plotpoints > ntime'
      end if
      timedata = ntime / plotpoints
      plotpoints = ntime / timedata
      write (*,*) 'Plotting ', plotpoints, ' points, i.e., every ', timedata,  &
     &            ' time steps a point'

      Mmax = Jmax
      ncols = (Jmax+1) * (4*Jmax*(Jmax+2)+3) / 3
      write (*,*) 'Number of J,K,M states ', ncols

      allocate(Vmat(-2:2,neq,-Mmax:Mmax,-Kmax:Kmax),     &
     &          c2tmat(-2:2,neq,-Mmax:Mmax,-Kmax:Kmax),  &
     &          dmcoeff(neq,ncols),                      &
     &          work(neq,5))
      allocate(xpcenter(nxpulses), ai_par_curr(nxpulses), ai_perp_curr(nxpulses), ai_thermal_curr(nxpulses),  &
     &          ai_par(nxpulses), ai_perp(nxpulses), ai_thermal(nxpulses),                        &
     &          ai_par_prev(nxpulses), ai_perp_prev(nxpulses), ai_thermal_prev(nxpulses),         &
     &          ai_par_perp(nxpulses), ai_par_thermal(nxpulses), ai_perp_thermal(nxpulses))

      ! PulseEnergy: total pulse energy, in mJ
      ! conversion to hartree
      PulseEnergy = PulseEnergy / 1000.0D0 * JouleToHartree
      ! tau: temporal width, in picoseconds, of Gaussian laser pulse (FWHM)
      ! conversion to atomic units
      tau = tau * PicosecondTot0
      ! Delta: spatial width, in micrometer, of Gaussian laser pulse (FWHM)
      ! conversion to bohr
      Delta = Delta * micrometerToBohr
      Delz2 = Delta**2
      ! Delta: spatial width, in micrometer, of Gaussian x-ray pulse (FWHM)
      ! conversion to bohr
      xrayDelta = xrayDelta * micrometerToBohr
      xrayDelz2 = xrayDelta**2
      ! Gaussian x-ray pulse duration (FWHM)
      tau_xr = tau_xr * PicosecondTot0

      ! temperature: (rotational) temperature, in K
      ! conversion to hartree
      temperature = temperature * KelvinToHartree

      ! A,B: rotational constants, in 1/cm
      ! Note: a prolate symmetric rotor is assumed.
      ! conversion to hartree
      RotConstA = RotConstA * invCentimeterToHartree
      RotConstB = RotConstB * invCentimeterToHartree

      ! Time propagation interval conversion to a.u.
      time_beg = time_beg * PicosecondTot0
      time_end = time_end * PicosecondTot0

      ! Determine times x-ray pulses are centered on
      t0 = time_beg
      dt = (time_end - time_beg) / real(nxpulses-1, dblprec)
      write (*,*) 'Length of time step for cross correlation ', dt * t0ToPicosecond, 'ps'
      do i=1, nxpulses
        xpcenter(i) = t0
        t0 = t0 + dt
      end do

      ! Calculate factor that allows to convert the laser power to
      ! an intensity (on the laser axis, rho=0). This conversion factor,
      ! when multiplied by power in hartree/t0, yields intensity in
      ! atomic units (an atomic unit of intensity defined as
      ! hartree/t0/a0^2=6.43641d15 W/cm^2)
      ! 4*ln(2)/pi = 0.8825424
      ! 4*ln(2) = 2.7725887
      PtoI = 4.0D0 * log(2.0D0) / M_PI      !0.8825424d0
      PtoI = PtoI / Delz2
      ! Conversion from number of x-rays distribution to x-ray flux
      NtoF = 4.0D0 * log(2.0D0) / M_PI
      NtoF = PtoI / xrayDelz2

      ! To determine the initial thermal distribution of rotational
      ! states stop here and evaluate the decay of the contributions
      ! to the partition function with J.
      partitionFunc = rotPartitionFunc(temperature)
      if (runmode == 3) then
        write (*,*)
        write (*,*) 'Thermal distribution successfully determined!'
        stop 'alignmol: thermal distribution!'
      end if

      ! Calculate rotational coupling matrix and matrix of cos^2(theta)
      call rotCoupleCos2t

      ! Integrate equations of motion from t0 to tf
      write (*,*) 'Molecular dynamics starts at ', time_beg * t0ToPicosecond,  &
     &            ' ps and ends at ', time_end * t0ToPicosecond, ' ps'

      ! Initial conditions (before pulse), i.e., a thermal
      ! density matrix
      call initdmcoeff

      ! Initialize temporary variables for the trapezoid rule
      intR_par           = 0
      intR_perp          = 0
      intR_thermal       = 0
      intR_par_curr      = 0
      intR_perp_curr     = 0
      intR_thermal_curr  = 0
      intR_par_prev      = 0
      intR_perp_prev     = 0
      intR_thermal_prev  = 0
      ai_par(:)          = 0
      ai_perp(:)         = 0
      ai_thermal(:)      = 0
      ai_par_curr(:)     = 0
      ai_perp_curr(:)    = 0
      ai_thermal_curr(:) = 0
      ai_par_prev(:)     = 0
      ai_perp_prev(:)    = 0
      ai_thermal_prev(:) = 0

      ! Prepare propagation
      t0 = time_beg
      dt = (time_end - time_beg) / real(ntime, dblprec)      ! dt: length of time step
      write (*,*) 'Time steps for propagation ', dt * t0ToPicosecond, ' ps'

      ! Time evolution
      do timestep = 1, ntime
        tf = t0 + dt

        if (mod(timestep, 1000) == 1) then
          write (*,*) 'Time step ', timestep, ' t0 =', t0 * t0ToPicosecond, ' ps'
        end if

        ! Propagation of the rotational wavepacket of the molecule
        call propDensMat(t0, tf, dt)

        ! Compute the ratio of the x-ray absorption cross
        ! section for parallel and perpendicular laser and x-ray
        ! polarizations
        if (runmode == 2) then
          call crosssection(timestep, tf, dt)
        end if

        ! Determine the expectation value of the cos^2 theta operator
        call calcc2t(neq, ncols, dmcoeff, Jmax, Kmax, Mmax, c2tmat, cos2t,  &
     &         RotConstA, RotConstB, temperature, partitionFunc, tf)

        ! Write data at a select number of time steps
        if (mod(timestep, timedata) == 0) then
          write (92,*) tf * t0ToPicosecond, real(cos2t)
          if (runmode == 2) then
            write (93,*) tf * t0ToPicosecond, PtoI * laserpower(tf, tau, PulseEnergy) * intensity_AUtoSI,  &
     &           NtoF * xrayflux(tf, 0.0D0, tau_xr, Nxrays)
            write (94,*) tf * t0ToPicosecond, intR_par_perp, intR_par_thermal, intR_perp_thermal
            write (96,*) tf * t0ToPicosecond, abs(upper_R_pp), abs(upper_R_par_thermal),  &
     &           abs(upper_R_perp_thermal)
            ! Debug stuff
!             write (149,*) tf * t0ToPicosecond, real(intR_par), real(intR_perp),  &
!      &            real(intR_thermal)
          else
            write (93,*) tf * t0ToPicosecond, PtoI * laserpower(tf, tau, PulseEnergy) * intensity_AUtoSI
          end if
        end if

        ! End of time step
        t0 = tf
      end do

      if (runmode == 2) then
        ! Write cross correlation
        do i=1, nxpulses
          write (97,*) xpcenter(i) * t0ToPicosecond, real(ai_par_perp(i)),  &
     &          real(ai_par_thermal(i)), real(ai_perp_thermal(i))
        end do

        ! Write waveplate scan
        do i=0, 360
          theta = real(i, dblprec)
          thetarad = theta * degTorad
          ratio =         real(intR_par)   * cos(thetarad)**2
          ratio = ratio + real(intR_perp)  * sin(thetarad)**2
          ratio = ratio / real(intR_thermal)
          write (98,*) theta, ratio
        end do
      end if

      ! Clean up and exit gracefully
      deallocate(xpcenter, ai_par_curr, ai_perp_curr, ai_thermal_curr,  &
     &          ai_par, ai_perp, ai_thermal,                            &
     &          ai_par_prev, ai_perp_prev, ai_thermal_prev,             &
     &          ai_par_perp, ai_par_thermal, ai_perp_thermal)
      deallocate(Vmat, c2tmat, dmcoeff, work)

      write (*,*)
      write (*,*) 'You spin me ''round ''round ''round ''round yeah! [Strong Interaction, Les Horribles Cernettes]'


      contains

!
! kappa_par --- compute the cross section factor for parallel
!       polarizations
!

      real(dblprec) function kappa_par(J1, J2, K, M)
      implicit none

      integer J1, J2, J3, K, M, K3

      kappa_par  = 0.0D0
      do J3 = J1-1, J1+1
        kappa_par  = kappa_par +  &
     &            reduced_S(J1, J3) * mol_sc_S(J1, J3, K) * lab_sz_S(J1, J3, M)  &
     &          * reduced_S(J3, J2) * mol_sc_S(J3, J2, K) * lab_sz_S(J3, J2, M)
        if (dipRatio /= 0.0D0) then
          do K3 = K-1, K+1, 2
            kappa_par  = kappa_par + dipRatio  &
     &          * reduced_S(J1, J3) * mol_sa_S(J1, K, J3, K3) * lab_sz_S(J1, J3, M)  &
     &          * reduced_S(J3, J2) * mol_sa_S(J3, K3, J2, K) * lab_sz_S(J3, J2, M)
          end do
        end if
      end do
      end function kappa_par


!
! kappa_perp --- compute the cross section factor for perpendicular
!       polarizations
!

      real(dblprec) function kappa_perp(J1, J2, K, M)
      implicit none

      integer J1, J2, K, M, J3, M3, K3

      kappa_perp = 0.0D0
      do M3 = M-1, M+1, 2
        do J3 = J1-1, J1+1
          kappa_perp = kappa_perp +  &
     &            reduced_S(J1, J3) * mol_sc_S(J1, J3, K) * lab_sx_S(J1, M, J3, M3)  &
     &          * reduced_S(J3, J2) * mol_sc_S(J3, J2, K) * lab_sx_S(J3, M3, J2, M)
          if (dipRatio /= 0.0D0) then
            do K3 = K-1, K+1, 2
              kappa_perp  = kappa_perp + dipRatio  &
     &          * reduced_S(J1, J3) * mol_sa_S(J1, K, J3, K3) * lab_sx_S(J1, M, J3, M3)  &
     &          * reduced_S(J3, J2) * mol_sa_S(J3, K3, J2, K) * lab_sx_S(J3, M3, J2, M)
            end do
          end if
        end do
      end do
      end function kappa_perp


!
! rotPartitionFunc --- calculate the initial themic density matrix
!       of rotational states
!

      real(dblprec) function rotPartitionFunc(temperature)
      use ridgidrotor
      implicit none

      integer J, K, M
      real(dblprec) Zaux1, Zaux2, beta, temperature
      real(dblprec) statweight
      external statweight

      ! Calculate rotational partition function.
      ! A C3v molecule such as CF3Br is assumed. The nuclear spin of
      ! fluorine is 1/2
      beta = 1.0D0 / temperature
      rotPartitionFunc = 0.0D0
      do J = 0, Jmax
        Zaux2 = 0.0D0
        do K = -min(J, Kmax), min(J, Kmax)
          Zaux1 = RotConstB * real(J*(J+1), dblprec) + (RotConstA - RotConstB) * real(K**2, dblprec)
          Zaux1 = beta * Zaux1
          Zaux1 = exp(-Zaux1)
          Zaux1 = Zaux1 * statweight(J, K, M)
          Zaux2 = Zaux2 + Zaux1
        end do
        Zaux2 = real(2*J+1, dblprec) * Zaux2
        write(91,*) J, Zaux2
        rotPartitionFunc = rotPartitionFunc + Zaux2
      end do
      end function rotPartitionFunc


!
! rotCoupleCos2t --- calculate the rotational coupling matrix and
!       the matrix of cos^2(theta)
!

      subroutine rotCoupleCos2t
      use anglib
      use ridgidrotor
      implicit none

      integer K, M, i, l, q, init, indr
      real(dblprec) aux, coeff

      ! Calculate rotational coupling matrix and matrix of cos^2(theta)
      do K = -Kmax, Kmax
         do M = -Mmax, Mmax
            init = max(abs(M), abs(K))
            indr = 0
            do i = init, init+neq-1
               indr = indr + 1
               do q = -2, 2
                  l = i + q
                  ! J'  = i
                  ! J'' = l
!                   aux = dsqrt(dble(2*l+1) / dble(2*i+1))
                  coeff = real(2*l+1, dblprec) / real(2*i+1, dblprec)
                  aux =       cleb(2*l, 2*K, 2*2, 0, 2*i, 2*K)
                  aux = aux * cleb(2*l, 2*M, 2*2, 0, 2*i, 2*M)
                  aux = 2.0D0 * aux / 3.0D0
                  ! For a couple of values the coefficient becomes negative.
                  ! However, fortunately, the Clebsch-Gordan coefficient
                  ! vanishes in this case.
                  if (coeff < 0.0D0) then
                    if (aux /= 0.0D0) then
                      write (*,*) '2*l+1, 2*i+1, aux'
                      write (*,*) 2*l+1, 2*i+1, aux
                      stop 'alignmol: illegal weight for matrix of cos^2(theta)!'
                    else
                      coeff = abs(coeff)
                    end if
                  end if
                  aux = aux * sqrt(coeff)
                  Vmat(q,indr,M,K) = PolAni * aux
                  c2tmat(q,indr,M,K) = aux
               end do
               c2tmat(0,indr,M,K) = c2tmat(0,indr,M,K) + 1.0D0 / 3.0D0
            end do
         end do
      end do
      end subroutine rotCoupleCos2t


!
! initdmcoeff --- initial conditions (before pulse), i.e., a thermal
!       density matrix
!

      subroutine initdmcoeff
      use ridgidrotor
      implicit none

      integer J, K, M, i, init, indr, indc

      indc = 0
      do J = 0, Jmax
         do K = -min(J, Kmax), min(J, Kmax)
            do M = -J, J
               indc = indc + 1
               init = max(abs(M), abs(K))
               indr = 0
               do i = init, init+neq-1
                  indr = indr + 1
                  if (i==J) then
                     dmcoeff(indr,indc) = (1.0D0,0.0D0)
                  else
                     dmcoeff(indr,indc) = (0.0D0,0.0D0)
                  end if
               end do
            end do
         end do
      end do
      end subroutine initdmcoeff


!
! propDensMat ---propagate in time the density matrix
!

      subroutine propDensMat(t0, tf, dt)
      use ridgidrotor
      implicit none

      integer J, K, M, q, init, indc
      real(dblprec) t0, tc, tf, dt, norm

      ! Numerical integration of rate equations
      tc = 0.5D0 * (t0 + tf)
      indc = 0
      do J = 0, Jmax
        do K = -min(J, Kmax), min(J, Kmax)
          do M = -J, J
            indc = indc + 1
            init = max(abs(M), abs(K))
            call rk4(neq, dmcoeff(1,indc), t0, tc, tf, dt, work, K, M, init)
            ! Enforce norm conservation
            norm = 0.0D0
            do q = 1, neq
              norm = norm + abs(dmcoeff(q,indc))**2
            end do
            if (abs(norm) > 1.0D-12) then
              norm = 1.0D0 / sqrt(norm)
            else
              norm = 0.0D0
            end if
            do q = 1, neq
              dmcoeff(q,indc) = norm * dmcoeff(q,indc)
            end do
          end do
        end do
      end do
      end subroutine propDensMat


!
! crosssection --- Compute the ratio of the x-ray absorption cross
!       section for parallel and perpendicular laser and x-ray
!       polarizations
!

      subroutine crosssection(timestep, tf, dt)
      use ridgidrotor
      implicit none

      integer timestep, J, J1, J2, K, M, q1, q2, i, init, indc
      real(dblprec) tf, dt, beta, Zaux
      complex(dblprec) r_par, r_perp, r_thermal, r_thermal2,  &
     &          rho_dens_thermal, phase, rd

      complex(dblprec), allocatable :: rho_dens(:,:,:,:)

      real(dblprec) statweight
      external statweight

      beta = 1.0D0 / temperature

      ! Assemble the density matrix elements
      allocate(rho_dens(neq, neq, -Kmax:Kmax, -Mmax:Mmax))
      rho_dens(:,:,:,:) = 0
      ! q1 is the row angular momentum
      do q1 = 1, neq
        ! q2 is the column angular momentum
        do q2 = 1, neq
          indc = 0
          do J = 0, Jmax
            do K = -min(J, Kmax), min(J, Kmax)
              do M = -J, J
                indc = indc + 1
                init = max(abs(M), abs(K))

                ! Thermodynamic factor
                Zaux = RotConstB * real(J*(J+1)) + (RotConstA - RotConstB) * real(K**2)
                Zaux = beta * Zaux
                Zaux = exp(-Zaux) / partitionFunc
                Zaux = Zaux * statweight(J, K, M)
                ! The density matrix element rho^{K,M}_{J1,J2}
                rho_dens(q1, q2, K, M) = rho_dens(q1, q2, K, M) + dmcoeff(q1,indc)  &
     &                                      * cmplx(Zaux) * conjg(dmcoeff(q2,indc))

              end do
            end do
          end do
        end do
      end do

      r_par     = 0
      r_perp    = 0
!       r_thermal = 0
!       r_thermal2 = 0
      ! Calculate the parallel and perpendicular angular coefficients
      do K = -Kmax, Kmax
        do M = -Mmax, Mmax
          init = max(abs(M), abs(K))
          do J1 = init, init+neq-1
            do J2 = init, init+neq-1
              ! Interaction picture to Schroedinger picture factor
              phase = RotConstB * real(J1*(J1+1) - J2*(J2+1), dblprec)
              phase = (0.0D0,-1.0D0) * phase * cmplx(tf)
              phase = exp(phase)
              q1 = J1 - init + 1
              q2 = J2 - init + 1
              rd = phase * rho_dens(q1, q2, K, M)
              ! The parallel and perpendicular angular coefficients
              r_par   = r_par   + rd * cmplx(kappa_par (J2,J1,K,M))
              r_perp  = r_perp  + rd * cmplx(kappa_perp(J2,J1,K,M))
            end do

!             ! Thermal ensemble angular coefficient
!             Zaux = RotConstB * real(J1*(J1+1)) + (RotConstA - RotConstB) * real(K**2)
!             Zaux = beta * Zaux
!             Zaux = exp(-Zaux) / partitionFunc
!             Zaux = Zaux * statweight(J1, K, M)
!             rho_dens_thermal = cmplx(Zaux)
!             ! kappa_par and thus r_thermal vanish for J1 = 0, i.e., T = 0K
!             r_thermal = r_thermal + rho_dens_thermal * cmplx(kappa_par(J1,J1,K,M))
!             r_thermal2 = r_thermal2 + rho_dens_thermal * cmplx(kappa_perp(J1,J1,K,M))
          end do
        end do
      end do
      deallocate(rho_dens)

      ! Thermal ensemble ratio
      r_thermal = 0
      r_thermal2 = 0
      do J = 0, Jmax
        do K = -min(J, Kmax), min(J, Kmax)
          do M = -J, J
            Zaux = RotConstB * real(J*(J+1)) + (RotConstA - RotConstB) * real(K**2)
            Zaux = beta * Zaux
            Zaux = exp(-Zaux) / partitionFunc
            Zaux = Zaux * statweight(J, K, M)
            rho_dens_thermal = cmplx(Zaux)
            r_thermal = r_thermal + rho_dens_thermal * cmplx(kappa_par(J,J,K,M))
            r_thermal2 = r_thermal2 + rho_dens_thermal * cmplx(kappa_perp(J,J,K,M))
          end do
        end do
      end do

      if (abs(r_thermal - r_thermal2) > 1.0D12) then
        stop 'crosssection: r_thermal /= r_thermal2'
      end if

      ! Direct instantaneous ratios
      upper_R_pp           = real(r_par ) / real(r_perp    + 1.0D-10)
      upper_R_par_thermal  = real(r_par ) / real(r_thermal + 1.0D-10)
      upper_R_perp_thermal = real(r_perp) / real(r_thermal + 1.0D-10)

      ! Compute current "rate"
      intR_par_curr     = r_par     * cmplx(xrayflux(tf, 0.0D0, tau_xr, Nxrays))
      intR_perp_curr    = r_perp    * cmplx(xrayflux(tf, 0.0D0, tau_xr, Nxrays))
      intR_thermal_curr = r_thermal * cmplx(xrayflux(tf, 0.0D0, tau_xr, Nxrays))
      do i=1, nxpulses
! For testing the cross correlation code
!         r_par  = PtoI * laserpower(tf, tau, PulseEnergy)
!         r_perp = 0
! End for testing the cross correlation code
        ai_par_curr(i)     = r_par     * cmplx(xrayflux(tf, xpcenter(i), tau_xr, Nxrays))
        ai_perp_curr(i)    = r_perp    * cmplx(xrayflux(tf, xpcenter(i), tau_xr, Nxrays))
        ai_thermal_curr(i) = r_thermal * cmplx(xrayflux(tf, xpcenter(i), tau_xr, Nxrays))
      end do

      ! To apply the trapezoid rule one needs two points, i.e., an interval
      if (timestep > 1) then
        intR_par     = intR_par     + cmplx(dt / 2.0D0) * (intR_par_prev     + intR_par_curr    )
        intR_perp    = intR_perp    + cmplx(dt / 2.0D0) * (intR_perp_prev    + intR_perp_curr   )
        intR_thermal = intR_thermal + cmplx(dt / 2.0D0) * (intR_thermal_prev + intR_thermal_curr)
        do i=1, nxpulses
          ai_par(i)     = ai_par(i)     + cmplx(dt / 2.0D0) * (ai_par_prev(i)     + ai_par_curr(i)    )
          ai_perp(i)    = ai_perp(i)    + cmplx(dt / 2.0D0) * (ai_perp_prev(i)    + ai_perp_curr(i)   )
          ai_thermal(i) = ai_thermal(i) + cmplx(dt / 2.0D0) * (ai_thermal_prev(i) + ai_thermal_curr(i))
        end do
      end if
      intR_par_prev      = intR_par_curr
      intR_perp_prev     = intR_perp_curr
      intR_thermal_prev  = intR_thermal_curr
      ai_par_prev(:)     = ai_par_curr(:)
      ai_perp_prev(:)    = ai_perp_curr(:)
      ai_thermal_prev(:) = ai_thermal_curr(:)

      ! Form integrated ratios
      intR_par_perp     = real(intR_par  + 1.0D-10) / real(intR_perp    + 1.0D-10)
      intR_par_thermal  = real(intR_par  + 1.0D-10) / real(intR_thermal + 1.0D-10)
      intR_perp_thermal = real(intR_perp + 1.0D-10) / real(intR_thermal + 1.0D-10)
      do i=1, nxpulses
        ai_par_perp(i)     = real(ai_par(i)  + 1.0D-10) / real(ai_perp(i)    + 1.0D-10)
        ai_par_thermal(i)  = real(ai_par(i)  + 1.0D-10) / real(ai_thermal(i) + 1.0D-10)
        ai_perp_thermal(i) = real(ai_perp(i) + 1.0D-10) / real(ai_thermal(i) + 1.0D-10)
      end do

      ! Debug writeout
!       write (150,*) tf * t0ToPicosecond, real(r_par), real(r_perp), real(r_thermal), real(r_thermal2)
!       write (151,*) tf * t0ToPicosecond, real(intR_par_curr), real(intR_perp_curr), real(intR_thermal_curr)
      end subroutine crosssection
      end program alignmol
