!
! caesr --- complete atomic electronic structure for rate equations
! Christian Buth
! 03 August 2016
!
! Version 1.0.0: 06 Oct 2016 (Nitrogen, neon, and argon are correctly GRASPed and the Auger rate is computed)
! Version 1.1.0: 12 Oct 2016 (Fluorescence rates and photoionization cross sections are computed)
! Version 1.2.0: 26 Oct 2016 (Batch scripting, nicer tables)
!
!
! References
! ----------
!
! Fritzsche, S., 2012, The RATIP program for relativistic calculations of atomic transition, ionization and
!     recombination properties. Comput. Phys. Commun. 183, 1525–1559, http://dx.doi.org/10.1016/j.cpc.2012.02.016
!
! Jönsson, P., G. Gaigalas, J. Bieroń, C. Froese Fischer, and I. P. Grant, 2013, New version: Grasp2K relativistic
!     atomic structure package, Comput. Phys. Commun., 184, 2197–2203, http://dx.doi.org/10.1016/j.cpc.2013.02.016
!
! Jönsson, P., X. He, C. Froese Fischer, and I. P. Grant, 2007, The grasp2K relativistic atomic structure package,
!     Comput. Phys. Commun. 177, 597–622, http://dx.doi.org/10.1016/j.cpc.2007.06.002
!


      program caesr
      use, intrinsic :: iso_fortran_env
      use fortrantl
      use physconst
      implicit none

      integer, parameter :: maxconf = 5, &                                    ! Maximum number of nonrelativistic subshells accepted
     &                      maxrelconf = maxconf + 2, &                          ! Maximum number of relativistic subshells accepted
     &                      maxgrasplev = 99, &                        ! Maximum number of levels used to optimize to in GRASP MCDHF
     &                      EOLthreshEAL = 40, &   ! Threshold of the number of CSFs for switching from EOL to EAL type computations
     &                      maxback = 10, &              ! Maximum number of previous GRASP computations used to find start orbitals
     &                      leninp = 132, &                                                          ! Maximum length of input lines
     &                      maxphotolines = 20**2 * 30       ! Maximum number of photoionization lines in cross section computations

      type configuration
        logical conv                                                                                ! GRASP computation is converged
        integer cfg(maxconf), &                                                    ! Nonrelativistic atomic electronic configuration
     &          nlevels, &                                                                                 ! Number of energy levels
     &          ncsf, &                                                                    ! Number of configuration state functions
     &          nregular                               ! Number of configuration state functions that lead to regular level energies
        integer, allocatable :: J1(:), J2(:)                               ! Numerator and denominator of the total angular momentum
        real(dblprec), allocatable :: energy(:)                                                        ! Eigenenergies of the levels
      end type configuration

      type transition
        integer ini, &                                                                                       ! Initial configuration
     &          fin                                                                                            ! Final configuration
        real(dblprec) energy, &                                                                                  ! Transition energy
     &                width                                                                                            ! Decay width
      end type transition

      type, extends(transition) :: augertransition
      end type augertransition

      type, extends(transition) :: radiativetransition
        logical E1_M1E2                                      ! Lowest multipole for the transition .true. for E1, .false. for M1, E2
      end type radiativetransition

      type crosssection
        integer ini, &                                                                                       ! Initial configuration
     &          fin                                                                                            ! Final configuration
        real(dblprec) ionthresh                                                                        ! Ionization threshold energy
        real(dblprec), allocatable :: energy(:), &                                                             ! Photon energy in eV
     &                                cross(:)                                                              ! Cross section in cm**2
      end type crosssection

      logical runstructure, &                                             ! Run GRASP computation of the atomic electronic structure
     &        runaugerdecay, &                                                           ! Run computation of the Auger decay widths
     &        runfluorescence, &                                                     ! Run computation of the radiative decay widths
     &        runcrosssection                                                                ! Run computation of the cross sections

      character(len=2), parameter :: subshellnames(maxconf) = (/'1s', '2s', '2p', '3s', '3p'/)      ! Nonrelativistic subshell names
      character(len=3), parameter :: relsubshellnames(maxrelconf) = (/'1s ', '2s ', '2p-', '2p ', '3s ', '3p-', '3p '/)
                                                                                                       ! Relativistic subshell names
      character(len=7), parameter :: relsubshells(maxconf)  = (/'1s     ', '2s     ', '2p-  2p', '3s     ', '3p-  3p'/)
                                                                                    ! Nonrelativistic to relativistic subshell names

      integer, parameter :: uicsl = 100, &                                                  ! IO unit for reading the GRASP CSL file
     &                      uocsl = 101, &                                                  ! IO unit for writing the GRASP CSL file
     &                      ujspt = 102, &                                                       ! IO unit for the GRASP jsplit file
     &                      umcp  = 103, &                                                          ! IO unit for the GRASP mcp file
     &                      uerwf = 104, &                                                         ! IO unit for the GRASP erwf file
     &                      urscf = 105, &                                                         ! IO unit for the GRASP rscf file
     &                      urlci = 106, &                                                        ! IO unit for the RATIP relci file
     &                      uaugr = 107, &                                                        ! IO unit for the RATIP auger file
     &                      uinul = 108, &                                                 ! IO unit for reading the GRASP rwfn file
     &                      uonul = 109, &                                                 ! IO unit for writing the GRASP rwfn file
     &                      usum  = 110, &                                                            ! IO unit for the summary file
     &                      ulog  = 111, &                                                                ! IO unit for the log file
     &                      ucesd = 112, &                                                         ! IO unit for the RATIP cesd file
     &                      ureos = 113, &                                                             ! IO unit for RATIP reos file
     &                      uphot = 114                                                               ! IO unit for RATIP photo file

      integer, parameter :: maprelnon(maxrelconf)   = (/1, 2, 3, 3, 4, 5, 5/), &              ! Mapping of relativistic subshells on
     & &                                                                                                 ! nonrelativistic subshells
     &                      maprelshell(maxrelconf) = (/1, 2, 2, 2, 3, 3, 3/), &          ! Mapping of relativistic subshells on the
     & &                                                                                                  ! principal quantum number
     &                      mapkappa(maxrelconf)    = (/-1, -1, 1, -2, -1, 1, -2/)     ! Relativistic quantum number of relativistic

      integer i, j, w1, w2, &
     &        nconfig, &                                                                       ! Number of electronic configurations
     &        nshells, &                                                                               ! Number of electronic shells
     &        nsubshells, &                                    ! Number of nonrelativistic subshells of the atom in its ground state
     &        nrelsubshells, &                                       ! Number of relativistic shells of the atom in its ground state
     &        nauger, &                                                                                ! Number of Auger transitions
     &        nradiative, &                                                                        ! Number of radiative transitions
     &        ncross, &                                                                                   ! Number of cross sections
     &        atomicnumber, &                                                                                ! Charge of the nucleus
     &        nxsects, &                                          ! Number of energies for which cross sections are to be determined
     &        nlingrid, &                                                 ! Number of energies on the linear grid for cross sections
     &        currcnf(maxconf), &                                                              ! Current configuration in the making
     &        startcnf(maxconf), &                                                  ! Starting configuration, e.g., the neutral atom
     &        endcnf(maxconf), &                                           ! Stopping configuration, e.g., the electron bare nucleus
     &        firstaugerdecay, &                                                             ! First Auger transition to be computed
     &        lastaugerdecay, &                                                              ! Last  Auger transition to be computed
     &        firstfluorescence, &                                                       ! First radiative transition to be computed
     &        lastfluorescence, &                                                        ! Last  radiative transition to be computed
     &        firstcrosssection, &                                                              ! First cross section to be computed
     &        lastcrosssection                                                                  ! Last  cross section to be computed

      real(dblprec), parameter :: threshshift = 0.001D0, &           ! Energy shift in eV of the threshold energy for cross sections
     &              linenergy = 10.0D0                                   ! Energy extent in eV of the linear grid for cross sections

      real(dblprec) convorbs, &                                                     ! Minimum amount of orbital convergence required
                    maxxsectenergy                                    ! Maximum energy for which a cross section shall be determined

      type(configuration) c
      type(configuration), allocatable :: configs(:)                  ! The configurations of the atom in all relevant charge states

      type(augertransition), allocatable :: augers(:)              ! The Auger transitions of the atom in all relevant charge states

      type(radiativetransition), allocatable :: radiatives(:)  ! The radiative transitions of the atom in all relevant charge states

      type(crosssection), allocatable :: crosses(:)                   ! The cross sections of the atom in all relevant charge states

      namelist /caesrinput/ atomicnumber, nshells, nsubshells, nrelsubshells, startcnf, endcnf, &
     &                      nlingrid, nxsects, maxxsectenergy, convorbs, runstructure, runaugerdecay, &
     &                      runfluorescence, runcrosssection, firstaugerdecay, lastaugerdecay, &
     &                      firstfluorescence, lastfluorescence, firstcrosssection, lastcrosssection


      ! Initialization
      write(*,*) 'CAESR V1.2.0, Complete atomic electronic structure for rate equations'
      write(*,*) '(c) 2016 Christian Buth'
      write(*,*)

      ! Nitrogen
      atomicnumber = 7
      nshells = 2
      nsubshells = 3
      nrelsubshells = nsubshells + 1
      startcnf = (/ 2, 2, 3, 0, 0 /)
      endcnf = 0
      nlingrid = 1
      nxsects = 3
      maxxsectenergy = 10000.0D0
      convorbs = 1.0D-5
      runstructure    = .true.
      runaugerdecay   = .true.
      runfluorescence = .true.
      runcrosssection = .true.
      firstaugerdecay   = 1
      lastaugerdecay    = 0
      firstfluorescence = 1
      lastfluorescence  = 0
      firstcrosssection = 1
      lastcrosssection  = 0

      ! Neon
!      atomicnumber = 10
!      nshells = 2
!      nsubshells = 3
!      nrelsubshells = nsubshells + 1
!      startcnf = 0
!      endcnf = 0
!      startcnf(1) = 2
!      startcnf(2) = 2
!      startcnf(3) = 6
!      nlingrid = 10
!      nxsects = 30
!      maxxsectenergy = 10000.0D0
!      convorbs = 1.0D-5

      ! Argon Nora Berrah
!      atomicnumber = 18
!      nshells = 3
!      nsubshells = 5
!      nrelsubshells = nsubshells + 2
!      startcnf = 0
!      endcnf = 0
!      startcnf(1) = 2
!      startcnf(2) = 2
!      startcnf(3) = 6
!      startcnf(4) = 2
!      startcnf(5) = 6
!      endcnf(1) = 2
!      nlingrid = 10
!      nxsects = 30
!      maxxsectenergy = 10000.0D0
!      convorbs = 1.0D-5

      ! Bhalla Table I
!      atomicnumber = 18
!      nshells = 3
!      nsubshells = 5
!      nrelsubshells = nsubshells + 2
!      startcnf = (/ 2, 2, 6, 2, 6 /)
!      endcnf   = (/ 1, 0, 0, 2, 0 /)
!      nlingrid = 10
!      nxsects = 30
!      maxxsectenergy = 10000.0D0
!      convorbs = 1.0D-5

      ! Bhalla Table II
!      atomicnumber = 18
!      nshells = 3
!      nsubshells = 5
!      nrelsubshells = nsubshells + 2
!      startcnf = (/ 2, 2, 6, 0, 0 /)
!      endcnf   = (/ 1, 0, 0, 0, 0 /)
!      nlingrid = 1
!      nxsects = 3
!      maxxsectenergy = 10000.0D0
!      convorbs = 1.0D-5
!      runstructure    = .false.
!      runaugerdecay   = .true.
!      runfluorescence = .false.
!      runcrosssection = .false.
!      firstaugerdecay   = 1
!      lastaugerdecay    = 0
!      firstfluorescence = 1
!      lastfluorescence  = 0
!      firstcrosssection = 1
!      lastcrosssection  = 0

      ! Argon
!      atomicnumber = 18
!      nshells = 3
!      nsubshells = 5
!      nrelsubshells = nsubshells + 2
!      startcnf = (/ 2, 2, 6, 2, 6 /)
!      endcnf   = (/ 0, 0, 0, 0, 0 /)
!      nlingrid = 10
!      nxsects = 30
!      maxxsectenergy = 10000.0D0
!      convorbs = 2.0D-5

      open(unit=usum, file='caesr.sum', status='replace', action='write', form='formatted', access='sequential')
      write(usum,*) 'CAESR summary file'
      write(usum,*)
      open(unit=ulog, file='caesr.log', status='replace', action='write', form='formatted', access='sequential')
      write(ulog,*) 'CAESR log file'
      write(ulog,*)

!      read(*,nml=caesrinput)

      write(usum,'(A,I0)') 'Atomic number: ', atomicnumber
      write(usum,'(A,I0)') 'Number of electron shells: ', nshells
      write(usum,'(A,I0)') 'Number of nonrelativistic electron subshells: ', nsubshells
      write(usum,'(A,I0)') 'Number of relativistic electron subshells: ', nrelsubshells
      write(usum,'(A,*(I6))') 'Starting configuration: ', startcnf(1:nsubshells)
      write(usum,'(A,*(I6))') '  Ending configuration: ', endcnf(1:nsubshells)
      write(usum,'(A,I0)') 'Number of energies for cross sections: ', nxsects
      write(usum,'(A,I0)') '  energies on a linear grid: ', nlingrid
      write(usum,'(A,F8.2,A)') '  up to a maximum energy of ', maxxsectenergy, ' eV'
      write(usum,'(A,E9.3)') 'Self-consistency of orbitals: ', convorbs
      write(usum,'(A)', advance='no') 'Tasks to be performed: '
      if (runstructure) then
        write(usum,'(A)', advance='no') 'runstructure  '
      end if
      if (runaugerdecay) then
        write(usum,'(A,2(I0,A))', advance='no') 'runaugerdecay(', firstaugerdecay, '--', lastaugerdecay, ')  '
      end if
      if (runfluorescence) then
        write(usum,'(A,2(I0,A))', advance='no') 'runfluorescence(', firstfluorescence, '--', lastfluorescence, ')  '
      end if
      if (runcrosssection) then
        write(usum,'(A,2(I0,A))', advance='no') 'runcrosssection(', firstcrosssection, '--', lastcrosssection, ')'
      end if
      write(usum,*)

      ! Generate configurations
      nconfig = 1
      do i = 1, nsubshells
        nconfig = nconfig * (startcnf(i) - endcnf(i) + 1)
      end do
      allocate(configs(nconfig))
      i = 1 ; j = 1
      call genconfigs(configs, i, nsubshells, j, currcnf, startcnf, endcnf)

      ! Sort configurations
      do i=1, nconfig
        w1 = sum(configs(i)%cfg)
        do j=i+1, nconfig
          w2 = sum(configs(j)%cfg)
          ! Lower charge states first
          if (w1 < w2) then
            c = configs(i)
            configs(i) = configs(j)
            configs(j) = c
            w1 = w2
          ! Inner shells occupied first
          else if (w1 == w2) then
            if (compconfigs(configs(i), configs(j), 1)) then
              c = configs(i)
              configs(i) = configs(j)
              configs(j) = c
            end if
          end if
        end do
      end do

      ! Print configurations
      write(usum,'(3/,2(A,/),/)') 'Electron configurations:', '------------------------'
      write(usum,*) ' Configuration No      Electron No  ', (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'Subshell'
      write(usum,*) '                                        ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(usum,*) '-------------------------------------', ('------', j=1, nsubshells)
      do i=1, nconfig
        write(usum,'(A,I6)', advance='no') '    ', i
        write(usum,'(A,I3)', advance='no') '                 ', sum(configs(i)%cfg)
        write(usum,'(A,*(A,I2))') '       ', ('    ', configs(i)%cfg(j), j=1, nsubshells)
      end do
      write(usum,*) '-------------------------------------', ('------', j=1, nsubshells)

      ! Compute atomic electronic structure for all configurations
      if (runstructure) then
        call graspconfigs(configs)
      end if
      if (runstructure .or. runaugerdecay .or. runfluorescence .or. runcrosssection) then
        call loadlevels(configs)
      end if

      ! Compute Auger transition energy and width
      if (runaugerdecay) then
        call augerenergies(configs, augers, nauger)
        if (lastaugerdecay == 0) then
          lastaugerdecay = nauger
        end if
        call augerwidths(configs, augers, firstaugerdecay, lastaugerdecay)
      end if

      ! Compute radiative transition energy and width
      if (runfluorescence) then
        call radiativeenergies(configs, radiatives, nradiative)
        if (lastfluorescence == 0) then
          lastfluorescence = nradiative
        end if
        call radiativewidths(configs, radiatives, firstfluorescence, lastfluorescence)
      end if

      ! Compute cross section energies and values
      if (runcrosssection) then
        call crosssectionenergies(configs, crosses, ncross, nxsects, maxxsectenergy)
        if (lastcrosssection == 0) then
          lastcrosssection = ncross
        end if
        call crosssectionvalues(configs, crosses, firstcrosssection, lastcrosssection, nxsects, maxxsectenergy)
      end if

      ! Clean up
      if (allocated(augers)) then
        deallocate(augers)
      end if
      if (allocated(radiatives)) then
        deallocate(radiatives)
      end if
      if (allocated(crosses)) then
        do i=1, ncross
          if (allocated(crosses(i)%energy)) then
            deallocate(crosses(i)%energy)
          end if
          if (allocated(crosses(i)%cross)) then
            deallocate(crosses(i)%cross)
          end if
        end do
        deallocate(crosses)
      end if
      do i=1, nconfig
        if (allocated(configs(i)%energy)) then
          deallocate(configs(i)%energy)
        end if
        if (allocated(configs(i)%J1)) then
          deallocate(configs(i)%J1)
        end if
        if (allocated(configs(i)%J2)) then
          deallocate(configs(i)%J2)
        end if
      end do
      deallocate(configs)
      write(ulog,'(/A)') ' end CAESR log file'
      close(ulog)
      if (.not. runcrosssection) then
        write (usum,*)
      end if
      write (usum,'(A)') ' end CAESR summary file'
      close(usum)
      write (*,'(/A)') ' Veni, vidi, vici [Gaius Julius Caesar]'


      contains

!
! genconfigs --- generate configurations of the atom
!

      recursive subroutine genconfigs(configs, iconf, nsubshells, isubshell, currcnf, startcnf, endcnf)
      implicit none

      integer, intent(in) :: nsubshells, isubshell
      integer, intent(inout) :: iconf, currcnf(maxconf), startcnf(maxconf), endcnf(maxconf)
      type(configuration), intent(inout) :: configs(:)
      integer k

      if (isubshell > nsubshells) then
        configs(iconf)%conv = .false.
        configs(iconf)%cfg(1:nsubshells) = currcnf(1:nsubshells)
        configs(iconf)%cfg(nsubshells+1:maxconf) = 0
        configs(iconf)%nlevels = -1
        configs(iconf)%ncsf = -1
        configs(iconf)%nregular = -1
        iconf = iconf + 1
        return
      end if

      do k=startcnf(isubshell), endcnf(isubshell), -1
        currcnf(isubshell) = k
        call genconfigs(configs, iconf, nsubshells, isubshell+1, currcnf, startcnf, endcnf)
      end do
      end subroutine genconfigs


!
! compconfigs --- compare configurations of the atom
!

      recursive logical function compconfigs(c1, c2, isubshell) result(compc)
      implicit none

      integer, intent(in) :: isubshell
      type(configuration), intent(in) :: c1, c2

      if (c1%cfg(isubshell) == c2%cfg(isubshell)) then
        if (isubshell >= nsubshells) then
          compc = .false.
          return
        else
          compc = compconfigs(c1, c2, isubshell + 1)
          return
        end if
      else if (c1%cfg(isubshell) < c2%cfg(isubshell)) then
          compc = .true.
          return
      end if

      compc = .false.
      end function compconfigs


!
! error_stop --- performs an error stop with outputting the error message to the screen and the log file
!

      subroutine error_stop(err)
      implicit none

      character(len=*), intent(in) :: err

      write(error_unit,'(/2A)') 'ERROR STOP ', err
      write(ulog,'(/2A)') 'ERROR STOP ', err
      error stop 'CAESR has failed'
      end subroutine error_stop


!
! execute_command --- calls execute_command_line and checks for error conditions
!

      subroutine execute_command(cmd, err)
      implicit none

      character(len=*), intent(in) :: cmd, err
      integer exstat, comstat
      character(len=200) :: commsg

      call execute_command_line(command=cmd, exitstat=exstat, cmdstat=comstat, cmdmsg=commsg)

      if (comstat /= 0) then
        call error_stop(err // new_line('x') // 'ERROR STOP ' // trim(commsg))
      end if

      if (exstat /= 0) then
        call error_stop(err)
      end if
      end subroutine execute_command


!
! csl_jsplit --- compute csl
!

      subroutine csl_jsplit(c, extracorr)
      implicit none

      type(configuration), intent(inout) :: c
      logical, intent(in) :: extracorr

      ! Generate configuration state list
      open(unit=uocsl, file='csl.inp', status='replace', action='write', form='formatted', access='sequential')
      ! 'Nonrelativistic' scheme?
      write(uocsl, '(A)') 'y'
      ! Enter the list of nonrelativistic core subshells:
      write(uocsl, '(A)')
      ! Enter the list of nonrelativistic peel subshells:
      ! If a shell is empty one needs to omit it to avoid an error in mcp1
      do i=1, nsubshells
        if (c%cfg(i) /= 0) then
          write(uocsl, '(A)', advance='no') '  ' // subshellnames(i)
        end if
      end do
      write(uocsl, '(A)')
      ! Redefine default maximum nonrelativistic peel subshell occupations?
      write(uocsl, '(A)') 'n'
      ! Redefine default minimum nonrelativistic peel subshell occupations?
      write(uocsl, '(A)') 'n'
      ! Enter a nonrelativistic reference peel subshell configuration (null if done):
      ! If a shell is empty one needs to omit it to avoid an error in mcp1
      do i=1, nsubshells
        if (c%cfg(i) /= 0) then
          write(uocsl, '(A,I0,A)', advance='no') subshellnames(i) // '(', c%cfg(i), ') '
        end if
      end do
      write(uocsl, '(A)')
      ! Enter the J values for this configuration (null for all possible J values):
      write(uocsl, '(A)')
      ! Additional correlations
      if (extracorr) then
        ! Nitrogen and neon
        if (atomicnumber == 7 .or. atomicnumber == 10) then

          ! The 1s and 2s are both singly occupied
          if (c%cfg(1) == 1 .and. c%cfg(2) == 1) then
            ! Enter a nonrelativistic reference peel subshell configuration (null if done):
            ! If a shell is empty one needs to omit it to avoid an error in mcp1
            if (c%cfg(3) > 1) then
              write(uocsl, '(A,I0,A)') '1s(2) 2p(', c%cfg(3), ') '
              write(ulog,'(A,I0,A)') ' Correlation configuration 1s(2) 2s(0) 2p(', c%cfg(3),')'
            else if (c%cfg(3) == 1) then
              write(uocsl, '(A)') '1s(2) 2p(1)'
              write(ulog,*) 'Correlation configuration 1s(2) 2s(0) 2p(1)'
            else if (c%cfg(3) == 0) then
              write(uocsl, '(A)') '1s(2)'
              write(ulog,*) 'Correlation configuration 1s(2) 2s(0) 2p(0)'
            end if
            ! Enter the J values for this configuration (null for all possible J values):
            write(uocsl, '(A)')
          end if

        ! Argon
        else if (atomicnumber == 18) then

          ! The 1s and 2s are both singly occupied
          if (c%cfg(1) == 1 .and. c%cfg(2) == 1) then
            ! Enter a nonrelativistic reference peel subshell configuration (null if done):
            ! If a shell is empty one needs to omit it to avoid an error in mcp1
            write(uocsl, '(A)', advance='no') '1s(2) '
            do i=3, nsubshells
              if (c%cfg(i) /= 0) then
                write(uocsl, '(A,I0,A)', advance='no') subshellnames(i) // '(', c%cfg(i), ') '
              end if
            end do
            write(uocsl, '(A)')
            write(ulog, '(A,3(I0,A))') ' Correlation configuration 1s(2) 2s(0) 2p(', &
     &                                 c%cfg(3),') 3s(', c%cfg(4),') 3p(', c%cfg(5),')'
            ! Enter the J values for this configuration (null for all possible J values):
            write(uocsl, '(A)')
          ! The 1s and 3s are both singly occupied
          else if (c%cfg(1) == 1 .and. c%cfg(4) == 1) then
            ! Enter a nonrelativistic reference peel subshell configuration (null if done):
            ! If a shell is empty one needs to omit it to avoid an error in mcp1
            write(uocsl, '(A)', advance='no') '1s(2) '
            do i=2, 3
              if (c%cfg(i) /= 0) then
                write(uocsl, '(A,I0,A)', advance='no') subshellnames(i) // '(', c%cfg(i), ') '
              end if
            end do
            if (c%cfg(5) /= 0) then
              write(uocsl, '(A,I0,A)', advance='no') '3p(', c%cfg(5),')'
            end if
            write(uocsl, '(A)')
            write(ulog, '(A,3(I0,A))') ' Correlation configuration 1s(2) 2s(', &
     &                                 c%cfg(2),') 2p(', c%cfg(3),') 3s(0) 3p(', c%cfg(5),')'
            ! Enter the J values for this configuration (null for all possible J values):
            write(uocsl, '(A)')
          ! The 2s and 3s are both singly occupied
          else if (c%cfg(2) == 1 .and. c%cfg(4) == 1) then
            ! Enter a nonrelativistic reference peel subshell configuration (null if done):
            ! If a shell is empty one needs to omit it to avoid an error in mcp1
            if (c%cfg(1) /= 0) then
              write(uocsl, '(A,I0,A)', advance='no') '1s(', c%cfg(1),') '
            end if
            write(uocsl, '(A)', advance='no') '2s(2) '
            if (c%cfg(3) /= 0) then
              write(uocsl, '(A,I0,A)', advance='no') '2p(', c%cfg(3),') '
            end if
            if (c%cfg(5) /= 0) then
              write(uocsl, '(A,I0,A)', advance='no') '3p(', c%cfg(5),')'
            end if
            write(uocsl, '(A)')
            write(ulog, '(A,3(I0,A))') ' Correlation configuration 1s(', &
     &                                 c%cfg(1),') 2s(2) 2p(', c%cfg(3),') 3s(0) 3p(', c%cfg(5),')'
            ! Enter the J values for this configuration (null for all possible J values):
            write(uocsl, '(A)')
          end if
          ! The 2p and 3p orbitals are equally occupied but neither empty nor full
          if (c%cfg(3) == c%cfg(5) .and. c%cfg(3) /= 0 .and. c%cfg(3) /= 6) then
            ! Enter a nonrelativistic reference peel subshell configuration (null if done):
            ! If a shell is empty one needs to omit it to avoid an error in mcp1
            do i=1, 2
              if (c%cfg(i) /= 0) then
                write(uocsl, '(A,I0,A)', advance='no') subshellnames(i) // '(', c%cfg(i), ') '
              end if
            end do
            write(uocsl, '(A,I0,A)', advance='no') '2p(', c%cfg(3) + 1, ') '
            if (c%cfg(4) /= 0) then
              write(uocsl, '(A,I0,A)', advance='no')'3s(', c%cfg(4), ') '
            end if
            write(uocsl, '(A,I0,A)', advance='no') '3p(', c%cfg(5) - 1,')'
            write(uocsl, '(A)')
            write(ulog, '(A,5(I0,A))') ' Correlation configuration 1s(', &
     &                  c%cfg(1),') 2s(', c%cfg(2),') 2p(', c%cfg(3) + 1,') 3s(', c%cfg(4), ') 3p(', c%cfg(5) - 1,')'
            ! Enter the J values for this configuration (null for all possible J values):
            write(uocsl, '(A)')
          end if

        ! Unknown atom
        else
          call error_stop('csl_jsplit: unknown atom')
        end if

      end if
      ! Enter a nonrelativistic reference peel subshell configuration (null if done):
      write(uocsl, '(A)')
      ! Enter the maximum number of substitutions permitted (0, ..., 9):
      write(uocsl, '(A)') '0'
      close(uocsl)
      call execute_command('$GRASP/bin/csl <csl.inp', 'csl_jsplit: error executing csl')

      ! Eliminate duplicate CSFs
      call execute_command('mv -fv rcsl.out rcsl.inp', 'csl_jsplit: error moving rcsl.out to rcsl.inp')
      open(unit=ujspt, file='jsplit.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Perform duplicate check and remove them ?
      write(ujspt, '(A)') 'y'
      close(ujspt)
      call execute_command('$GRASP/bin/jsplit <jsplit.inp', 'csl_jsplit: error executing jsplit')
      end subroutine csl_jsplit


!
! csfs_mcp --- compute csl and angular integration
!

      subroutine csfs_mcp(c, extracorr)
      implicit none

      type(configuration), intent(inout) :: c
      logical, intent(in) :: extracorr

      ! Generate CSL of the multiplett
      if (.not. extracorr) then
        call csl_jsplit(c, extracorr)
        c%ncsf = ncsf_corr(.true.)
        c%nregular = c%ncsf
        write(ulog,'(A,I0)') ' Number of CSFs ', c%ncsf
      ! Generate CSL with additional correlation effects
      else
        call csl_jsplit(c, extracorr)
        c%nregular = ncsf_corr(.false.)
        write(ulog,'(2(A,I0))') ' Number of CSFs plain ', c%ncsf, ', with additional correlations ', c%nregular
      end if

      ! The number of levels used in MCDHF must not exceed maxgrasplev;  otherwise rscf1 terminates with
      ! the error message "CONVRT: Length of CNUM inadeuate"
      c%nregular = min(c%nregular, maxgrasplev)
      call execute_command('cp -fv rcsl.out rcsl.inp', 'csfs_mcp: error copying rcsl.out to rcsl.inp')

      ! Angular integration
      open(unit=umcp, file='mcp.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Default settings?
      write(umcp, '(A)') 'y'
      close(umcp)
      call execute_command('$GRASP/bin/mcp1 <mcp.inp', 'csfs_mcp: error executing mcp1')
      end subroutine csfs_mcp


!
! ncsf_corr --- determine the number of CSFs in a CSL file and correct the CSL file for RATIP
!

      integer function ncsf_corr(ratipcsl)
      implicit none

      logical, intent(in) :: ratipcsl
      character(len=leninp) line
      integer i, ncsf

      ! Read and correct GRASP CSL file for RATIP
      ncsf = -huge(1)
      open(unit=uicsl, file='rcsl.out', status='old', action='read', form='formatted', access='sequential')
      if (ratipcsl) then
        open(unit=uocsl, file='rcsl_0.out', status='replace', action='write', form='formatted', access='sequential')
      end if
      do
        read(uicsl, '(A)', end=10) line
        ! Insert missing orbitals
        if (trim(line) == 'Peel subshells:') then
          if (ratipcsl) then
            write(uocsl, '(A)') trim(line)
            read(uicsl, '(A)')
            do i=1, nsubshells
              write(uocsl, '(A)', advance='no') '  ' // trim(relsubshells(i)) // ' '
            end do
            write(uocsl, '(A)')
            end if
          cycle
        ! Count the CSFs
        else if (trim(line) == 'CSF(s):') then
          ncsf = 0
        end if
        ! Skip GRASP block delimiters
        if (scan(line, '*') == 0) then
          if (ratipcsl) then
            write(uocsl, '(A)') trim(line)
          end if
          ncsf = ncsf + 1
        end if
      end do
 10   if (ratipcsl) then
        close(uocsl)
      end if
      close(uicsl)

      ncsf_corr = ncsf / 3
      end function ncsf_corr


!
! erwf_TF --- estimate starting wave function for GRASP based on the Thomas-Fermi approximation
!

      subroutine erwf_TF()
      implicit none

      open(unit=uerwf, file='erwf.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Default settings ?
      write(uerwf, '(A)') 'y'
      ! Read subshell radial wavefunctions. Choose one below
      !     1 -- GRASP92 File
      !     2 -- Thomas-Fermi
      !     3 -- Screened Hydrogenic
      write(uerwf, '(A)') '2'
      ! Enter the list of relativistic subshells:
      write(uerwf, '(A)') '*'
      close(uerwf)
      call execute_command('$GRASP/bin/erwf <erwf.inp', 'erwf_TF: error executing erwf')
      end subroutine erwf_TF


!
! erwf_GRASP --- estimate starting wave function for GRASP based on a previous converged computation
!

      subroutine erwf_GRASP(c)
      implicit none

      type(configuration), intent(in) :: c
      character(len=leninp) line
      integer j

      ! Use rwfn.out as starting orbitals
      write(line, '(A,*(A,I0))') 'structure/rscf', ('_', c%cfg(j), j = 1, nsubshells)
      write(ulog,*) 'Attempting starting orbitals from ', trim(line)
      call execute_command('cp -fv ' // trim(line) // '/rwfn.out .', 'erwf_GRASP: error copying rwfn.out to the current directory')

      ! Estimating relativistic wave functions
      open(unit=uerwf, file='erwf.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Default settings ?
      write(uerwf, '(A)') 'y'
      ! Read subshell radial wavefunctions. Choose one below
      !     1 -- GRASP92 File
      !     2 -- Thomas-Fermi
      !     3 -- Screened Hydrogenic
      write(uerwf, '(A)') '1'
      ! Enter the file name (Null then "rwfn.out")
      write(uerwf, '(A)') 'rwfn.out'
      ! Enter the list of relativistic subshells:
      write(uerwf, '(A)') '*'
      ! Read subshell radial wavefunctions. Choose one below
      !     1 -- GRASP92 File
      !     2 -- Thomas-Fermi
      !     3 -- Screened Hydrogenic
      write(uerwf, '(A)') '2'
      ! Enter the list of relativistic subshells:
      write(uerwf, '(A)') '*'
      close(uerwf)
      call execute_command('$GRASP/bin/erwf <erwf.inp', 'erwf_GRASP: error executing erwf')
      end subroutine erwf_GRASP


!
! mcdhf --- carry out MCDHF computation and check for its convergence
!

      recursive logical function mcdhf(c, EOL, exceeded) result(mcdhfres)
      implicit none

      type(configuration), intent(inout) :: c
      logical, intent(in) :: EOL
      logical, intent(out) :: exceeded
      logical csfreduce

      ! Multi configuration Dirac Hartree Fock computation
      open(unit=urscf, file='rscf.inp', status='replace', action='write', form='formatted', access='sequential')
      ! EOL computation
      if (EOL) then
        write (ulog,*) 'Extended optimal level [EOL] type computation'
        ! Default settings?
        write(urscf, '(A)') 'y'
        ! EOL type calculation?
        write(urscf, '(A)') 'y'
        ! Enter the serial numbers of the ASF(s)
        write(urscf, '(A,I0)') '1-', c%nregular
        ! Enter orbitals to be varied (Updating order)
        write(urscf, '(A)') '*'
        ! Which of these are spectroscopic orbitals?
        write(urscf, '(A)') '*'
        ! Enter the maximum number of SCF cycles:
        write(urscf, '(A)') '100'

      ! EAL computation
      else
        write (ulog,*) 'Extended average level [EAL] type computation'
        ! Default settings?
        write(urscf, '(A)') 'y'
        ! EOL type calculation?
        write(urscf, '(A)') 'n'
        ! EAL type calculation; H(DC) will not be diagonalised;
        ! Standard ASF weights are (2J+1)
        !  values; revise?
        write(urscf, '(A)') 'n'
        ! Enter the maximum number of SCF cycles:
        write(urscf, '(A)') '100'
      end if
      close(urscf)

      call execute_command('rm -fv rwfn.out', 'mcdhf: error removing rwfn.out')
      call execute_command('$GRASP/bin/rscf1 <rscf.inp >rscf.stdout 2>rscf.stderr', 'mcdhf: error excecuting rscf1')
      call execute_command('cat >rscf.output rscf.stdout rscf.stderr', &
     &                     'mcdhf: error concatenating rscf.stdout and rscf.stderr to rscf.output')
      call execute_command('cat rscf.output', 'mcdhf: error outputting rscf.output')

      mcdhfres = check_mcdhf(c, exceeded, csfreduce)

      if (csfreduce) then
        write(ulog,*) 'mcdhf: recursive call with nregular ', c%nregular
        mcdhfres = mcdhf(c, EOL, exceeded)
      end if
      end function mcdhf


!
! check_mcdhf --- check whether an MCDHF computation was performed correctly
!

      logical function check_mcdhf(c, exceeded, csfreduce)
      implicit none

      type(configuration), intent(inout) :: c
      logical, intent(out) :: exceeded, &
     &                        csfreduce
      logical fileexists, &                                                                      ! Test whether a file exists or not
     &        levforopt
      character(len=leninp) line                                                                         ! Line of input from a file
      integer i
      real(dblprec) levelenergy

      ! Check correctness and convergence of the rscf computation
      csfreduce = .false.
      levforopt = .false.
      open(unit=urscf, file='rscf.output', status='old', action='read', form='formatted', access='sequential')
      do
        read(urscf, '(A)', end=11) line
        if (trim(line) == '  Maximum iterations in SCF Exceeded.') then
          write(ulog,*) 'Not converged: maximum iterations in SCF exceeded'
          close(urscf)
          check_mcdhf = .false.
          exceeded = .true.
          return
        else if (trim(line) == ' IN: maximum tabulation point exceeds') then
          write(ulog,*) 'Not converged: IN: maximum tabulation point exceeds'
          close(urscf)
          check_mcdhf = .false.
          exceeded = .false.
          return
        else if (trim(line) == ' Convergence not obtained') then
          write(ulog,*) 'Convergence not obtained'
          close(urscf)
          check_mcdhf = .false.
          exceeded = .false.
          return
        else if (trim(line) == ' Not a sorted GRASP92 MCP File;') then
          call error_stop('check_mcdhf: not a sorted GRASP92 MCP File')
        else if (trim(line) == ' Failure; equation for orbital') then
          call error_stop('check_mcdhf: failure; equation for orbital could not be solved')
        else if (trim(line) == ' CONVRT: Length of CNUM inadeuate.') then
          call error_stop('check_mcdhf: CONVRT: Length of CNUM inadeuate')
        ! Determine whether all level energies are regular
        else if (trim(line) == 'Optimise on the following level(s):' .and. .not. levforopt) then
          levforopt = .true.
          read(urscf,*)
          do i=1, c%nregular
            read(urscf, '(T22,E19.12)') levelenergy
            if (abs(levelenergy) > 1.0D20) then
              write (ulog,*) 'Singular level energy'
              close(urscf)
              c%nregular = i - 1
              check_mcdhf = .false.
              exceeded = .false.
              csfreduce = .true.
              return
            end if
          end do
        end if
      end do
 11   close(urscf)

      inquire(file='rwfn.out', exist=fileexists)
      if (.not. fileexists) then
        check_mcdhf = .false.
        exceeded = .false.
        return
      end if

      check_mcdhf = .true.
      exceeded = .false.
      end function check_mcdhf


!
! graspconfigs --- compute with GRASP checking for reduced but still acceptable orbital convergence
!

      subroutine graspconfigs(configs)
      implicit none

      type(configuration), intent(inout) :: configs(:)
      logical exceeded, extracorr, EOL
      integer icnf, j, tryconf

      ! Iterate over all atomic configurations
      do icnf=1, nconfig
        configs(icnf)%conv = .false.

        ! Do not GRASP an electron-bare configuration
        if (sum(configs(icnf)%cfg) == 0) then
          write(ulog,'(3/A,*(I6))') ' Electron-bare configuration (not GRASPed)', icnf, (configs(icnf)%cfg(j), j=1, nsubshells)
          cycle
        end if

        if (icnf > 1) then
          write(*,'(2/)')
          write(ulog,'(2/)')
        end if
        write(*,'(A,*(I6))') 'GRASPing ', icnf, (configs(icnf)%cfg(j), j=1, nsubshells)
        write(ulog,'(A,*(I6))') ' GRASPing ', icnf, (configs(icnf)%cfg(j), j=1, nsubshells)

        ! Generate CSFs and perform angular integration
        extracorr = .false.
        call csfs_mcp(configs(icnf), extracorr)
        EOL = configs(icnf)%ncsf <= EOLthreshEAL

        ! Estimating relativistic orbitals with Thomas-Fermi approximation
 10     write(ulog,*) 'Estimating relativistic orbitals with Thomas-Fermi approximation'
        call erwf_TF()
        configs(icnf)%conv = mcdhf(configs(icnf), EOL , exceeded)
        if (configs(icnf)%conv) then
          go to 12
        end if

        if (exceeded) then
          if (orbitalconv()) then
            configs(icnf)%conv = .true.
            go to 12
          end if
        end if

        ! Estimating relativistic orbitals with previous GRASP computations
        write(ulog,'(/A)') ' Estimating relativistic orbitals with previous GRASP computations'
        do tryconf=icnf-1, max(1, icnf - maxback), -1
          ! Only try converged configurations
          if (.not. configs(tryconf)%conv) then
            cycle
          end if
          call erwf_GRASP(configs(tryconf))

          ! Try to converge the multiplet
          configs(icnf)%conv = mcdhf(configs(icnf), EOL, exceeded)
          if (configs(icnf)%conv) then
            go to 12
          end if
          if (exceeded) then
            if (orbitalconv()) then
              configs(icnf)%conv = .true.
              go to 12
            end if
          end if
        end do

        ! Failed to GRASP the configuration
 11     if (extracorr) then
          write(ulog,*) 'graspconfigs: tried all strategies...failure for configuration ', &
     &                  (configs(icnf)%cfg(j), j=1, nsubshells)
          cycle
        end if

        ! Generate CSFs and perform angular integration
        write (ulog,*)
        extracorr = .true.
        call csfs_mcp(configs(icnf), extracorr)
        EOL = configs(icnf)%nregular <= EOLthreshEAL
        if (configs(icnf)%ncsf == configs(icnf)%nregular) then
          go to 11
        else
          go to 10
        end if

 12     call orbitals_relci_cesd(configs(icnf))
      end do
      end subroutine graspconfigs


!
! orbitalconv --- check orbital convergence of an rscf1 computation
!

      logical function orbitalconv(suben)
      implicit none

      character(len=leninp) line                                                                         ! Line of input from a file
      integer j, irelsubs
      real(dblprec), intent(out), optional :: suben(nrelsubshells)
      real(dblprec) rm1, r, r2, genocc, selfcon, maxselfcon

      open(unit=urscf, file='rscf.sum', status='old', action='read', form='formatted', access='sequential')
      read(urscf, '(4/,A11)', advance='no') line
      read(urscf,*) irelsubs

      do
        read(urscf, '(A)') line

        ! Section on relativistic orbital energies
        if (present(suben) .and. trim(line) == &
     &      ' Subshell           e                    p0                  gamma' // &
     &      '                   P(2)                  Q(2)          MTP') then
          read(urscf,*)
          suben = 0.0D0
          do j=1, irelsubs
            read(urscf, '(A12)', advance='no') line
            read(urscf, *) suben(j)
          end do

        ! Section on self consistency
        else if (trim(line) == &
     & ' Subshell       <  r     >             <  r  >              <  r    >             occupation           consistency') then
          read(urscf,*)
          maxselfcon = 0.0D0
          do j=1, irelsubs
            read(urscf, '(A12)', advance='no') line
            read(urscf, *) rm1, r, r2, genocc, selfcon
            if (selfcon > maxselfcon) then
              maxselfcon = selfcon
            end if
          end do
          close(urscf)

          write(ulog,*) 'Orbital convergence equal or better than ', maxselfcon
          if (maxselfcon > convorbs) then
            orbitalconv = .false.
          else
            orbitalconv = .true.
          end if
          return
        end if

      end do
      end function orbitalconv


!
! orbitals_relci_cesd --- compute CI matrix from CSFs with Breit interaction
!

      subroutine orbitals_relci_cesd(c)
      implicit none

      type(configuration), intent(in) :: c

      logical fileexists, &                                                                      ! Test whether a file exists or not
              generatednull                                              ! .true. if a null orbitals was generated otherwise .false.

      character(len=6) G92RWF                                                     ! Indicator of a GRASP92 relativistic orbital file
      character(len=leninp) line                                                                         ! Line of input from a file

      integer j, &
     &        np, &                                                                       ! Principal quantum number of the electron
     &        nak, &                                                                  ! Angular quantum number kappa of the electron
     &        mfj, &                                                            ! The number of points to represent radial functions
     &        numorb, &                                                                                ! Relativistic orbital number
     &        inro                                 ! Nonrelativistic orbital number corresponding to the relativistic orbital number

      real(dblprec) e, &                                                                               ! The energy of the electron
     &              pz                                                    ! Zero-order expansion coefficient of the radial function

      real(dblprec), allocatable :: PF(:), &                                           ! The large component of the radial function
     &                              QF(:), &                                           ! The small component of the radial function
     &                              R(:)                                                                    ! The radial coordinate

      ! Read and correct GRASP rwfn file.  Generate null orbitals for missing rscf1 orbitals.
      ! RATIP performs a normalization and a Schmidt orthogonalization of the orbitals.  This needs to be commented
      ! out for null orbitals to work.  They are on lines 950-955 and line 980 in rabs_grasp92.f90.
      ! The file format of the rwfn file is described on page 606 of (Jönsson et al., 2007)
      inquire(file='rwfn_0.out', exist=fileexists)
      if (fileexists) then
        call execute_command('rm -fv rwfn_0.out', 'orbitals_relci_cesd: error removing rwfn_0.out')
      end if
      open(unit=uinul, file='rwfn.out', status='old', action='read', form='unformatted', access='sequential')
      open(unit=uonul, file='rwfn_0.out', status='replace', action='write', form='unformatted', access='sequential')
      read(uinul) G92RWF
      if (G92RWF /= 'G92RWF') then
        call error_stop('orbitals_relci_cesd: not a GRASP92 relativistic wave function file!')
      end if
      write(uonul) G92RWF

      ! Iterate over all relativistic radial functions
      generatednull = .false.
      write (ulog,'(/A/)') ' Subshell           E                    p0             mfj'
      do numorb=1, nrelsubshells
        inro = maprelnon(numorb)
        if (c%cfg(inro) /= 0) then
          ! Copy relativistic radial function from rwfn.out to rwfn_0.out
          read(uinul) np, nak, e, mfj
          write(uonul) np, nak, e, mfj
          allocate (PF(mfj), QF(mfj), R(mfj))
          read(uinul) pz, PF(:), QF(:)
          read(uinul) R(:)
          write(uonul) pz, PF(:), QF(:)
          write(uonul) R(:)
          deallocate (PF, QF, R)
          write (ulog,'(3A,2(E18.12,A),I0)') '    ', relsubshellnames(numorb), '     ', e, '    ', pz, '    ', mfj

        ! Generate null orbital
        else
          np = maprelshell(numorb)
          nak = mapkappa(numorb)
          e = 0.0D0
          ! mfj needs to be 13 or larger for the null orbital to be accepted
          mfj = 13
          write(uonul) np, nak, e, mfj
          allocate (PF(mfj), QF(mfj), R(mfj))
          pz = 0.0D0 ; PF = 0.0D0 ; QF = 0.0D0
          R(1) = 0.0D0
          do j=2, mfj
            R(j) = 0.1D0 * real(j, dblprec)
          end do
          write(uonul) pz, PF(:), QF(:)
          write(uonul) R(:)
          deallocate (PF, QF, R)
          generatednull = .true.
        end if
      end do
      close(uonul)
      close(uinul)

      ! If no null orbital is required then replace rwfn_0.out by a symbolic link to rwfn.out
      if (.not. generatednull) then
        call execute_command('rm -fv rwfn_0.out', 'orbitals_relci_cesd: error removing rwfn_0.out')
        call execute_command('ln -sv rwfn.out rwfn_0.out', &
     &                       'orbitals_relci_cesd: error creating a symbolic link from rwfn_0.out to rwfn.out')
      else
         write (ulog,'(/A)') ' Null orbitals were emitted for missing orbitals'
      end if

      ! RELCI computation of the multiplet
      inquire(file='relci.sum', exist=fileexists)
      if (fileexists) then
        call execute_command('rm -fv relci.sum', 'orbitals_relci_cesd: error removing relci.sum')
      end if
      inquire(file='relci.mix', exist=fileexists)
      if (fileexists) then
        call execute_command('rm -fv relci.mix', 'orbitals_relci_cesd: error removing relci.mix')
      end if
      open(unit=urlci, file='relci.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Enter a file name for the  relci.sum  file:
      write(urlci, '(A)') 'relci.sum'
      ! Enter the name of the configuration symmetry list file:
      write(urlci, '(A)') 'rcsl_0.out'
      ! Enter the name of the isotope data file:
      write(urlci, '(A)') 'isodata'
      ! Modify default set-up and printout of the program ?
      write(urlci, '(A)') 'y'
      ! Include contributions of the frequency-independent Breit interaction ?
      write(urlci, '(A)') 'y'
      ! Include vacuum polarization contributions to H ?
      write(urlci, '(A)') 'y'
      ! Include specific mass shift contributions to H ?
      write(urlci, '(A)') 'y'
      ! Estimate contributions from self-energy ?
      write(urlci, '(A)') 'n'
      ! Store the effective interaction strengths in memory ?
      !  This should always be true for a on-fly calculation of the Hamiltonian matrix.
      write(urlci, '(A)') 'y'
      ! Precalculate and keep Hamiltonian matrix in memory ?
      write(urlci, '(A)') 'y'
      ! Diagonalize the full Hamiltonian matrix ?
      ! (I.e. independent of the parameter hamiltonian_fullmatrix.)
      write(urlci, '(A)') 'y'
      ! Which units are to be used to enter and to print the energies of the continuum orbitals ?
      !     A       : Angstrom;
      !     eV      : electron volts;
      !     Hartree : Hartree atomic units;
      !     Hz      : Hertz;
      !     Kayser  : [cm**(-1)];
      write(urlci, '(A)') 'eV'
      ! The physical speed of light in atomic units is   137.03598950000000      ;
      !  revise this value ?
      write(urlci, '(A)') 'n'
      ! The default radial grid parameters for this case are:
      !  rnt =    ....................... ;
      !  h   =    ....................... ;
      !  hp  =    ....................... ;
      !  n   =         .... ;
      !  revise these values ?
      write(urlci, '(A)') 'n'
      ! Enter the name of the GRASP92 Radial WaveFunction File:
      write(urlci, '(A)') 'rwfn_0.out'
      ! Enter a file name for the  relci.mix  file:
      write(urlci, '(A)') 'relci.mix'
      ! Enter the serial number(s) of the level(s) to be calculated;
      !  e.g. 1 3 4  7 - 20  48  69 - 85;
      write(urlci, '(A,I0)') '1-', c%ncsf
      close(urlci)
      call execute_command('$RATIP/xrelci <relci.inp', 'orbitals_relci_cesd: error executing xrelci')

      ! CESD computation of the multiplet
      inquire(file='cesd.sum', exist=fileexists)
      if (fileexists) then
        call execute_command('rm -fv cesd.sum', 'orbitals_relci_cesd: error removing cesd.sum')
      end if
      inquire(file='cesd.xpn', exist=fileexists)
      if (fileexists) then
        call execute_command('rm -fv cesd.xpn', 'orbitals_relci_cesd: error removing cesd.xpn')
      end if
      open(unit=ucesd, file='cesd.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Enter a file name for the  cesd.sum  file:
      write(ucesd, '(A)') 'cesd.sum'
      ! Enter the name of the configuration symmetry list file:
      write(ucesd, '(A)') 'rcsl_0.out'
      ! Expand the atomic state functions (ASF) ?
      write(ucesd, '(A)') 'y'
      ! Print non-standard output on the cesd expansion ?
      write(ucesd, '(A)') 'n'
      ! Select non-standard total M values for the  expansion into Slater determinants ?
      !  The standard is to use M = +J for each individual determinant.
      write(ucesd, '(A)') 'n'
      ! Enter the name of the GRASP92 mixing
      write(ucesd, '(A)') 'relci.mix'
      ! Enter a name for the CESD eXPaNsion  .xpn  output File that is to be created:
      write(ucesd, '(A)') 'cesd.xpn'
      close(ucesd)
      call execute_command('$RATIP/xcesd <cesd.inp', 'orbitals_relci_cesd: error executing xcesd')

      ! Store files from GRASP and RELCI computations
      inquire(file='structure', exist=fileexists)
      if (.not. fileexists) then
        call execute_command('mkdir -v structure', 'orbitals_relci_cesd: cannot create directory structure')
      end if
      write(line, '(A,*(A,I0))') 'structure/rscf', ('_', c%cfg(j), j = 1, nsubshells)
      inquire(file=trim(line), exist=fileexists)
      if (.not. fileexists) then
        call execute_command('mkdir -v ' // trim(line), 'orbitals_relci_cesd: cannot create directory')
      end if
      call execute_command('mv -vf rcsl.out rcsl_0.out rscf.sum rwfn.out rwfn_0.out relci.sum relci.mix cesd.sum cesd.xpn ' &
     &                     // trim(line), 'orbitals_relci_cesd: not all files of rcsl.out, rcsl_0.out, rscf.sum, rwfn.out,' &
     &                     // ' rwfn_0.out, relci.sum, relci.mix cesd.sum cesd.xpn found!')
      inquire(file='rmix.out', exist=fileexists)
      if (fileexists) then
      call execute_command('mv -vf rmix.out ' // trim(line), 'orbitals_relci_cesd: rmix.out not found!')
      end if
      call execute_command('rm -fv fort.* mcp.* sms.20 *.inp rscf.stdout rscf.stderr rscf.output', &
     &                     'orbitals_relci_cesd: error removing fort.*, mcp.*, sms.20, *.inp, rscf.stdout ' &
     &                     // 'rscf.stderr rscf.output')
      end subroutine orbitals_relci_cesd


!
! loadlevels --- load level energies from RELCI computations
!

      subroutine loadlevels(configs)
      implicit none

      type(configuration), intent(inout) :: configs(:)
      logical fileexists
      character(len=leninp) line                                                                         ! Line of input from a file
      integer icfg, i, j, k, nlevels
      real(dblprec) r, eigenenergy, sf, sumsf, energy

      ! Load level energies
      write(usum,'(3/,2(A,/),/)') 'Energy levels:', '--------------'
      write(usum,*) '   Config.   Levels       Lowest level     Highest level        Average' &
     &           // '          Spread   ', (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'Subshell'
      write(usum,*) '                              [eV]              [eV]              [eV] ' &
     &           // '           [eV]        ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(usum,*) '-------------------------------------------------------------------------------------------', &
     &             ('------', j=1, nsubshells)
      do icfg=1, nconfig
        write (line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(icfg)%cfg(j), j = 1, nsubshells)
        inquire(file=trim(line), exist=fileexists)
        configs(icfg)%conv = fileexists
        if (.not. fileexists) then
          cycle
        end if
        open(unit=urlci, file=trim(line) // '/relci.sum', status='old', action='read', form='formatted', access='sequential')

        ! Read the number of levels
        nlevels = 1
        do
          read(urlci, '(A)', end=10) line
          if (line(1:11) == ' A total of') then
            read(line(12:24),*) nlevels
            exit
          end if
        end do
        ! The string ' A total of' appears only if there is more than one level
 10     if (nlevels == 1) then
          rewind(urlci)
        end if
        configs(icfg)%nlevels = nlevels

        ! Read the eigenenergies
        do
          read(urlci, '(A)') line
          if (trim(line) == ' Eigenenergies:') then
            read(urlci, '(2/)')
            allocate(configs(icfg)%energy(nlevels), configs(icfg)%J1(nlevels), configs(icfg)%J2(nlevels))
            do k=1, nlevels
              read(urlci, '(A20)', advance='no') line
              read(urlci,*) r, eigenenergy
              configs(icfg)%energy(k) = eigenenergy
              ! Read total angular momentum
              read(line(10:13),*) configs(icfg)%J1(k)
              if (line(12:12) == '/') then
                read(line(13:14),*) configs(icfg)%J2(k)
              else
                configs(icfg)%J2(k) = 1
              end if
            end do
            exit
          end if
        end do
        close(urlci)

        ! Output energies
        sumsf = 0.0D0
        energy = 0.0D0
        do i=1, configs(icfg)%nlevels
          sf = 2.0D0 * real(configs(icfg)%J1(i), dblprec) / real(configs(icfg)%J2(i), dblprec) + 1.0D0
          energy = energy + sf * configs(icfg)%energy(i)
          sumsf = sumsf + sf
        end do
        energy = energy / sumsf
        write(usum, '(2(A,I6))', advance='no') '  ', icfg, '   ', configs(icfg)%nlevels
        write(usum, '(3(A,F15.6),A,F12.6)', advance='no') '       ', configs(icfg)%energy(1), '   ', &
     &                   configs(icfg)%energy(configs(icfg)%nlevels), '   ', energy, &
     &                   '   ', configs(icfg)%energy(configs(icfg)%nlevels) - configs(icfg)%energy(1)
        write(usum, '(A, *(I6))') ' ', (configs(icfg)%cfg(j), j=1, nsubshells)
      end do
      write(usum,*) '-------------------------------------------------------------------------------------------', &
     &             ('------', j=1, nsubshells)
      end subroutine loadlevels


!
! augerenergies --- generate Auger transitions and compute energies
!

      subroutine augerenergies(configs, augers, nauger)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(augertransition), allocatable, intent(out) :: augers(:)
      integer, intent(out) :: nauger
      logical finex, fingtini
      integer ini, fin, k, l, p, diffcfg(maxconf)
      real(dblprec) augerenergy, inisf, finsf, sf, inien, finen
      type(augertransition), allocatable :: tmpaug(:)

      ! Identify Auger channels and compute Auger energies
      write(ulog,'(3/,2(A/),/)') 'Auger transitions:', '------------------'
      write(ulog,*) '   Aug. trn.   In. cfg.   Fi. cfg.       Energy     ', &
     &              (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &              (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
      write(ulog,*) '                                          [eV]       ' &
     &           // '   ', (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      allocate(augers(1))
      nauger = 0
      do ini=1, nconfig
 trans: do fin=ini+1, nconfig
          ! Check the difference of the configurations for potential Auger transitions
          diffcfg = configs(ini)%cfg - configs(fin)%cfg
          l = 0 ; p = 1
          do k=1, nsubshells
            ! Only nonzero elements matter
            if (diffcfg(k) /= 0) then
              l = l + 1
              ! More than three nonzero elements?  Then this is no Auger transition
              if (l > 3) then
                cycle trans
              ! The final state has a lower subshell with an electron more than the initial state
              else if (l == 1 .and. diffcfg(k) /= -1) then
                cycle trans
              ! Store the index of the second nonzero element
              else if (l == 2) then
                p = k
              ! The final state has two electrons less either in one upper subshell or distributed over two upper subshells
              else if (l == 2 .and. diffcfg(k) /= 1 .and. diffcfg(k) /= 2) then
                cycle trans
              ! If there is a third nonzero element, then it needs to be 1
              else if (l == 3 .and. diffcfg(k) /= 1) then
                cycle trans
              ! If the previous nonzero element is not 1 this is no Auger transition
              else if (l == 3 .and. diffcfg(k) == 1 .and. diffcfg(p) /= 1) then
                cycle trans
              end if
            end if
          end do
          ! If there is no nonzero element or just one nonzero element, then this is no Auger transition either
          if (l == 0 .or. l == 1) then
            cycle trans
          ! If there are only two nonzero elements, the first must be -1 and the second 2
          else if (l == 2 .and. diffcfg(p) /= 2) then
            cycle trans
          end if

          ! Auger transitions occur if there are initial states with a higher energy than the final states
          if (.not. configs(ini)%conv .or. .not. configs(fin)%conv) then
            write(ulog, '(A,*(I3))', advance='no') ' Incomplete atomic electronic structure from ', &
     &                                               (configs(ini)%cfg(k), k=1, nsubshells)
            write(ulog, '(A,*(I3))') ' to ', (configs(fin)%cfg(k), k=1, nsubshells)
            cycle trans
          end if
          if (maxval(configs(ini)%energy) >= minval(configs(fin)%energy)) then
            nauger = nauger + 1

            ! Probabilistic average over initial state multiplet
            inien = 0.0D0
            inisf = 0.0D0
            fingtini = .false.
            do i=1, configs(ini)%nlevels

              ! Probabilistic average over final state multiplet
              finen = 0.0D0
              finsf = 0.0D0
              finex = .false.
              do j=1, configs(fin)%nlevels
                ! Only include final state energy levels that are lower in energy than the initial state level
                if (configs(ini)%energy(i) > configs(fin)%energy(j)) then
                  sf = 2.0D0 * real(configs(fin)%J1(j), dblprec) / real(configs(fin)%J2(j), dblprec) + 1.0D0
                  finen = finen + sf * configs(fin)%energy(j)
                  finsf = finsf + sf
                  finex = .true.
                else
                  if (.not. fingtini) then
                    write(ulog, *) 'Final-state energy with larger energy than the initial state:'
                  end if
                  fingtini = .true.
                end if
              end do

              ! Sum up level energy difference, if there is a final state
              if (finex) then
                finen = finen / finsf
                sf = 2.0D0 * real(configs(ini)%J1(i), dblprec) / real(configs(ini)%J2(i), dblprec) + 1.0D0
                inien = inien + sf * (configs(ini)%energy(i) - finen)
                inisf = inisf + sf
              end if
            end do
            augerenergy = inien / inisf

            write(ulog, '(3(A,I6))', advance='no') '   ', nauger, '     ', ini, '     ', fin
            write(ulog, '(A,F15.6)', advance='no') '     ', augerenergy
            write(ulog, '(A,*(I6))', advance='no') '  ', (configs(ini)%cfg(k), k=1, nsubshells)
            write(ulog, '(A,*(I6))')               '  ', (configs(fin)%cfg(k), k=1, nsubshells)

            ! Store Auger transition
            if (nauger > 1) then
              call move_alloc(augers, tmpaug)
              allocate(augers(nauger))
              augers(1:nauger-1) = tmpaug(:)
              deallocate(tmpaug)
            end if

            augers(nauger)%ini    = ini
            augers(nauger)%fin    = fin
            augers(nauger)%energy = augerenergy
            augers(nauger)%width  = -1.0D0
          end if
        end do trans
      end do
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      end subroutine augerenergies


!
! augerwidths --- compute widths of Auger transitions
!

      subroutine augerwidths(configs, augers, firstaugerdecay, lastaugerdecay)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(augertransition), allocatable, intent(inout) :: augers(:)
      integer, intent(in) :: firstaugerdecay, lastaugerdecay
      logical fileexists                                                                         ! Test whether a file exists or not
      character(len=leninp) line, str                                                                    ! Line of input from a file
      integer i, j, k, l, iaug, levnum
      real(dblprec) augwidth, levwidth, sumsf, sf

      ! Create directory for xauger files
      inquire(file='augerdecay', exist=fileexists)
      if (.not. fileexists) then
        call execute_command('mkdir -v augerdecay', 'augerwidths: cannot create directory augerdecay')
      end if

      ! Compute width of all Auger transitions
      write(ulog,'(3/2(A/))') 'Auger widths:', '-------------'
      write(usum,'(3/,2(A/),/)') 'Auger transitions:', '------------------'
      write(usum,*) '   Aug. trn.   In. cfg.   Fi. cfg.       Energy            Width      ', &
     &              (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &              (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
      write(usum,*) '                                          [eV]             [meV]       ' &
     &           // '   ', (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(usum,*) '-------------------------------------------------------------------------', &
     &              ('------', j=1, 2*nsubshells)
      do iaug=firstaugerdecay, lastaugerdecay
        write(ulog, '(I6,A,*(I3))', advance='no') iaug, ' Auger transition from ', &
     &              (configs(augers(iaug)%ini)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,*(I3))', advance='no') ' to ', &
     &              (configs(augers(iaug)%fin)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,F15.6,A)') ' energy: ', augers(iaug)%energy, ' eV'

        ! Gather files for xauger computation
        inquire(file='auger.sum', exist=fileexists)
        if (fileexists) then
          call execute_command('rm -vf auger.sum', 'augerwidths: unable to remove auger.sum')
        end if
        write (line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(augers(iaug)%ini)%cfg(j), j = 1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/rcsl_0.out rcsl1.out', &
     &                       'augerwidths: unable to copy rcsl_0.out to rcsl1.out')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn1.out', &
     &                       'augerwidths: unable to copy rwfn_0.out to rwfn1.out')
        call execute_command('cp -fv ' // trim(line) // '/relci.mix relci1.mix', &
     &                       'augerwidths: unable to copy relci.mix to relci1.mix')
        write (line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(augers(iaug)%fin)%cfg(j), j = 1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/rcsl_0.out rcsl2.out', &
     &                       'augerwidths: unable to copy rcsl_0.out to rcsl2.out')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn2.out', &
     &                       'augerwidths: unable to copy rwfn_0.out to rwfn2.out')
        call execute_command('cp -fv ' // trim(line) // '/relci.mix relci2.mix', &
     &                       'augerwidths: unable to copy relci.mix to relci2.mix')

        ! Compute Auger transition with xauger
        open(unit=uaugr, file='auger.inp', status='replace', action='write', form='formatted', access='sequential')
        ! Enter a file name for the  auger.sum  file:
        write(uaugr, '(A)') 'auger.sum'
        ! Enter the name of the isotope data file:
        write(uaugr, '(A)') 'isodata'
        ! Which units are to be used to enter and to print the energies of the continuum orbitals ?
        !     A       : Angstrom;
        !     eV      : electron volts;
        !     Hartree : Hartree atomic units;
        !     Hz      : Hertz;
        !     Kayser  : [cm**(-1)];
        write(uaugr, '(A)') 'eV'
        ! Enter the maximal energy of the Auger transitions  (in eV) to built-up the radial grid;
        write(uaugr, '(I0)') ceiling(maxval(configs(augers(iaug)%ini)%energy) - minval(configs(augers(iaug)%fin)%energy) + 50.0D0)
        ! Modify default set-up and printout of the program ?
        write(uaugr, '(A)') 'y'
        ! Select individual transitions ?
        write(uaugr, '(A)') 'n'
        ! Include exchange interactions into the generation of the continuum waves ?
        write(uaugr, '(A)') 'y'
        ! Include Breit interactions to the Auger matrix ?
        write(uaugr, '(A)') 'y'
        ! Calculate angular distribution parameters ?
        write(uaugr, '(A)') 'n'
        ! Calculate spin polarization parameters ?
        write(uaugr, '(A)') 'n'
        ! Sort transitions in ascending order of energy ?
        write(uaugr, '(A)') 'y'
        ! Read in and apply experimental energies for the calculation of Auger rates and other properties ?
        write(uaugr, '(A)') 'n'
        ! Auger rates are printed in SI units; use Hartree atomic units instead ?
        write(uaugr, '(A)') 'n'
        ! Print all selected transitions and their energies before the computation starts (this is the default) ?
        write(uaugr, '(A)') 'y'
        ! Print the CSF scheme each time a new one has been built ?
        write(uaugr, '(A)') 'n'
        ! Print the results for each individual transition immediatly after its computation ?
        write(uaugr, '(A)') 'n'
        ! Print the final results to a (.trn) transition amplitude file ?
        write(uaugr, '(A)') 'n'
        ! Enter an (overall) shift of the Auger energies which applies to all transitions (ineV):
        !  Use  0.  or   <cr>  if no shift need to be applied.
        write(uaugr, '(A)') '0'
        ! Enter a minimal energy (> =0.) of the free electron (ineV):
        !  All other transitions are neglected from the computations;
        !  use  0.  or   <cr>  if all possible transitions are to be taken into account.
        write(uaugr, '(A)') '0'
        ! Enter a maximal (-)kappa symmetry up to which continuum spinors are taken into account ?
        !  2 (up to p-waves), 4(f), 6(i), 8(k), ...; 0 or <cr> to include all possible waves.
        write(uaugr, '(A)') '0'
        ! The physical speed of light in atomic units is   137.03598950000000      ;
        !  revise this value ?
        write(uaugr, '(A)') 'n'
        ! The default radial grid parameters for this case are:
        !  rnt =    ....................... ;
        !  h   =    ....................... ;
        !  hp  =    ....................... ;
        !  n   =         .... ;
        !  revise these values ?
        write(uaugr, '(A)') 'n'
        ! Enter the name of the initial-state GRASP92 configuration symmetry list file:
        write(uaugr, '(A)') 'rcsl1.out'
        ! Enter the name of the final-state GRASP92 configuration symmetry list file:
        write(uaugr, '(A)') 'rcsl2.out'
        ! Enter the name of the initial-state GRASP92 mixing coefficient file:
        write(uaugr, '(A)') 'relci1.mix'
        ! Enter the name of the final-state GRASP92 mixing coefficient file:
        write(uaugr, '(A)') 'relci2.mix'
        ! Enter the name of the initial-state Radial WaveFunction File:
        write(uaugr, '(A)') 'rwfn1.out'
        ! Enter the name of the final-state Radial WaveFunction File:
        write(uaugr, '(A)') 'rwfn2.out'
        close(uaugr)
        call execute_command('$RATIP/xauger <auger.inp', 'augerwidths: xauger returned an error')

        ! Extract Auger width
        open(unit=uaugr, file='auger.sum', status='old', action='read', form='formatted', access='sequential')
        do
          read(uaugr, '(A)') line
          if (trim(line) == ' Auger lifetimes, total rates and widths :') then
            read(uaugr, '(6/)')
            augwidth = 0.0D0
!            sumsf = 0.0D0
            i = 1
            do l=1, configs(augers(iaug)%ini)%nlevels
              read(uaugr, '(A)') line
              if (trim(line) == &
     & ' -------------------------------------------------------------------------------------------------------------------') &
     &        then
                exit
              end if
              read(line, '(I7)') levnum
              read(line, '(T101,E13.7)') levwidth
              sf = 2.0D0 * real(configs(augers(iaug)%ini)%J1(levnum), dblprec) &
     &                   / real(configs(augers(iaug)%ini)%J2(levnum), dblprec) + 1.0D0
              augwidth = augwidth + sf * levwidth
!              sumsf = sumsf + sf
              i = l
            end do
            ! Test whether a level has zero decay width
            if (i < configs(augers(iaug)%ini)%nlevels) then
              write(ulog,*) '      Level with zero Auger decay width present'
            end if
            ! Sum up the statistical factor.  If a level has zero decay width it is omitted in the above sum.
            ! Thus the sum needs to be computed separately
            sumsf = 0.0D0
            do l=1, configs(augers(iaug)%ini)%nlevels
              sf = 2.0D0 * real(configs(augers(iaug)%ini)%J1(l), dblprec) &
     &                   / real(configs(augers(iaug)%ini)%J2(l), dblprec) + 1.0D0
              sumsf = sumsf + sf
            end do
            augwidth = augwidth / sumsf
            augers(iaug)%width = augwidth

            ! Print Auger transition
            write(usum, '(3(A,I6))', advance='no') '   ', iaug, '     ', augers(iaug)%ini, '     ', augers(iaug)%fin
            write(usum, '(A,F15.6,A,E13.6)', advance='no') '     ', augers(iaug)%energy, '     ', augers(iaug)%width*1000.0D0
            write(usum, '(A,*(I6))', advance='no') '  ', (configs(augers(iaug)%ini)%cfg(k), k=1, nsubshells)
            write(usum, '(A,*(I6))')               '  ', (configs(augers(iaug)%fin)%cfg(k), k=1, nsubshells)
            exit
          end if
        end do
        close(uaugr)

        ! Treat files from the xauger computation
        write(line, '(A,*(A,I0))') 'auger', ('_', configs(augers(iaug)%ini)%cfg(j), j = 1, nsubshells)
        write(str, '(A,*(A,I0))') '_to',    ('_', configs(augers(iaug)%fin)%cfg(j), j = 1, nsubshells)
        line = trim(line) // trim(str) // '.sum'
        call execute_command('mv -vf auger.sum augerdecay/' // trim(line), &
     &                       'augerwidths: unable to move auger.sum to augerdecay/' // trim(line))
        call execute_command('gzip -vf9 augerdecay/' // trim(line), &
     &                       'augerwidths: unable to compress augerdecay/' // trim(line))
        call execute_command('rm -vf auger.inp rcsl1.out rcsl2.out rwfn1.out rwfn2.out relci1.mix relci2.mix', &
     &                       'augerwidths: unable to remove auger.inp rcsl1.out rcsl2.out rwfn1.out ' // &
     &                       'rwfn2.out relci1.mix relci2.mix')
      end do
      write(usum,*) '-------------------------------------------------------------------------', &
     &              ('------', j=1, 2*nsubshells)
      end subroutine augerwidths


!
! radiativeenergies --- generate radiative transitions and compute energies
!

      subroutine radiativeenergies(configs, radiatives, nradiative)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(radiativetransition), allocatable, intent(out) :: radiatives(:)
      integer, intent(out) :: nradiative
      logical E1_M1E2, finex, fingtini
      integer ini, fin, i, j, k, l, diffcfg(maxconf)
      real(dblprec) radiativeenergy, inisf, finsf, sf, inien, finen
      type(radiativetransition), allocatable :: tmprad(:)

      ! Identify radiative channels and compute radiative energies
      write(ulog,'(3/,2(A/),/)') 'Radiative transitions:', '----------------------'
      write(ulog,*) '   Rad. trn.   In. cfg.   Fi. cfg.       Energy     ', &
     &              (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &              (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
      write(ulog,*) '                                          [eV]       ' &
     &           // '   ', (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      allocate(radiatives(1))
      nradiative = 0
      do ini=1, nconfig
 trans: do fin=1, nconfig
          ! Check the difference of the configurations for potential radiative transitions
          diffcfg = configs(ini)%cfg - configs(fin)%cfg
          l = 0 ; i = 0; j = 0
          do k=1, nsubshells
            ! Only nonzero elements matter
            if (diffcfg(k) /= 0) then
              l = l + 1
              ! More than two nonzero elements?  Then this is no radiative transition
              if (l > 2) then
                cycle trans
              ! The final state has a lower subshell with an electron more than the initial state
              else if (l == 1 .and. diffcfg(k) /= -1) then
                cycle trans
              ! Store the value of the lower shell
              else if (l == 1) then
                i = k
              ! The initial state has an upper subshell with an electron more than the final state
              else if (l == 2 .and. diffcfg(k) /= 1) then
                cycle trans
              ! Store the value of the upper subshell
              else if (l == 2) then
                j = k
              end if
            end if
          end do
          ! If there is no nonzero element then this is no radiative transition either
          if (l /= 2) then
            cycle trans
          end if

          ! Identify the lowest multipole order of the radiative transition
          if (atomicnumber > 18) then
            call error_stop('radiativeenergies: unknown atom')
          end if
          E1_M1E2 = .not. (((i == 1 .or. i == 2) .and. (j == 2 .or. j == 4)) .or. (i == 3 .and. j == 5))

          ! Radiative transitions occur if there are initial states with a higher energy than the final states
          if (.not. configs(ini)%conv .or. .not. configs(fin)%conv) then
            write(ulog, '(A,*(I3))', advance='no') ' Incomplete atomic electronic structure from ', &
     &                                              (configs(ini)%cfg(k), k=1, nsubshells)
            write(ulog, '(A,*(I3))') ' to ', (configs(fin)%cfg(k), k=1, nsubshells)
            cycle trans
          end if
          if (maxval(configs(ini)%energy) >= minval(configs(fin)%energy)) then
            nradiative = nradiative + 1

            ! Probabilistic average over initial state multiplet
            inien = 0.0D0
            inisf = 0.0D0
            fingtini = .false.
            do i=1, configs(ini)%nlevels

              ! Probabilistic average over final state multiplet
              finen = 0.0D0
              finsf = 0.0D0
              finex = .false.
              do j=1, configs(fin)%nlevels
                ! Only include final state energy levels that are lower in energy than the initial state level
                if (configs(ini)%energy(i) > configs(fin)%energy(j)) then
                  sf = 2.0D0 * real(configs(fin)%J1(j), dblprec) / real(configs(fin)%J2(j), dblprec) + 1.0D0
                  finen = finen + sf * configs(fin)%energy(j)
                  finsf = finsf + sf
                  finex = .true.
                else
                  if (.not. fingtini) then
                    write(ulog, *) 'Final-state energy with larger energy than the initial state:'
                  end if
                  fingtini = .true.
                end if
              end do

              ! Sum up level energy difference, if there is a final state
              if (finex) then
                finen = finen / finsf
                sf = 2.0D0 * real(configs(ini)%J1(i), dblprec) / real(configs(ini)%J2(i), dblprec) + 1.0D0
                inien = inien + sf * (configs(ini)%energy(i) - finen)
                inisf = inisf + sf
              end if
            end do
            radiativeenergy = inien / inisf

            write(ulog, '(3(A,I6))', advance='no') '   ', nradiative, '     ', ini, '     ', fin
            write(ulog, '(A,F15.6)', advance='no') '     ', radiativeenergy
            write(ulog, '(A,*(I6))', advance='no') '  ', (configs(ini)%cfg(k), k=1, nsubshells)
            write(ulog, '(A,*(I6))')               '  ', (configs(fin)%cfg(k), k=1, nsubshells)

            ! Store radiative transition
            if (nradiative > 1) then
              call move_alloc(radiatives, tmprad)
              allocate(radiatives(nradiative))
              radiatives(1:nradiative-1) = tmprad(:)
              deallocate(tmprad)
            end if

            radiatives(nradiative)%E1_M1E2 = E1_M1E2
            radiatives(nradiative)%ini     = ini
            radiatives(nradiative)%fin     = fin
            radiatives(nradiative)%energy  = radiativeenergy
            radiatives(nradiative)%width   = -1.0D0
          end if
        end do trans
      end do
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      end subroutine radiativeenergies


!
! radiativewidths --- compute widths of radiative transitions
!

      subroutine radiativewidths(configs, radiatives, firstfluorescence, lastfluorescence)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(radiativetransition), allocatable, intent(inout) :: radiatives(:)
      integer, intent(in) :: firstfluorescence, lastfluorescence
      logical fileexists                                                                         ! Test whether a file exists or not
      character(len=leninp) line                                                                         ! Line of input from a file
      integer j, k, irad
      real(dblprec) radM1, radE2

      ! Create directory for xreos files
      inquire(file='fluorescence', exist=fileexists)
      if (.not. fileexists) then
        call execute_command('mkdir -v fluorescence', 'radiativewidths: cannot create directory fluorescence')
      end if

      ! Compute width of all radiative transitions
      write(ulog,'(3/,2(A/))') 'Radiative widths:', '-----------------'
      write(usum,'(3/,2(A/),/)') 'Radiative transitions:', '----------------------'
      write(usum,*) '   Rad. trn.   In. cfg.   Fi. cfg.       Energy            Width         Multipol  ', &
     &              (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &              (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
      write(usum,*) '                                          [eV]             [meV]                    ' &
     &           // '   ', (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(usum,*) '--------------------------------------------------------------------------------------', &
     &              ('------', j=1, 2*nsubshells)
      do irad=firstfluorescence, lastfluorescence
        write(ulog, '(I6,A,*(I3))', advance='no') irad, ' Radiative transition from ', &
     &              (configs(radiatives(irad)%ini)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,*(I3))', advance='no') ' to ', &
     &              (configs(radiatives(irad)%fin)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,F15.6,A)') ' energy: ', radiatives(irad)%energy, ' eV'

        ! Gather files for xreos computation
        inquire(file='reos.sum', exist=fileexists)
        if (fileexists) then
          call execute_command('rm -vf reos.sum', 'radiativewidths: unable to remove reos.sum')
        end if
        write(line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(radiatives(irad)%ini)%cfg(j), j = 1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/cesd.xpn cesd1.xpn', &
     &                       'radiativewidths: unable to copy cesd.xpn to cesd1.xpn')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn1.out', &
     &                       'radiativewidths: unable to copy rwfn_0.out to rwfn1.out')
        write(line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(radiatives(irad)%fin)%cfg(j), j = 1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/cesd.xpn cesd2.xpn', &
     &                       'radiativewidths: unable to copy cesd.xpn to cesd2.xpn')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn2.out', &
     &                       'radiativewidths: unable to copy rwfn_0.out to rwfn2.out')

        ! Carry out radiative width computation for a specific multipole
        if (radiatives(irad)%E1_M1E2) then
          radiatives(irad)%width = radiativemultipole(configs, radiatives(irad), 'E1')
        else
          radM1 = radiativemultipole(configs, radiatives(irad), 'M1')
          radE2 = radiativemultipole(configs, radiatives(irad), 'E2')
          radiatives(irad)%width = radM1 + radE2
        end if

        ! Print radiative transition
        if (radiatives(irad)%E1_M1E2) then
          write(ulog, '(A,E13.6,A)') '       E1 transition width:', radiatives(irad)%width*1000.0D0, ' meV'
        else
          write(ulog, '(2(A,E13.6),A)') '       M1 transition width: ', radM1*1000.0D0, &
     &                                   ' meV, E2 transition width: ', radE2*1000.0D0, ' meV'
        end if
!        write(ulog,*)

        write(usum, '(3(A,I6))', advance='no') '   ', irad, '     ', radiatives(irad)%ini, '     ', radiatives(irad)%fin
        write(usum, '(A,F15.6,A,E13.6)', advance='no') '     ', radiatives(irad)%energy, '     ', radiatives(irad)%width*1000.0D0
        if (radiatives(irad)%E1_M1E2) then
          write(usum, '(A)', advance='no') '        E1   '
        else if (radM1 /= 0.0D0 .and. radE2 /= 0.0D0) then
          write(usum, '(A)', advance='no') '      M1 + E2'
        else if (radM1 /= 0.0D0) then
          write(usum, '(A)', advance='no') '        M1   '
        else if (radE2 /= 0.0D0) then
          write(usum, '(A)', advance='no') '        E2   '
        end if
        write(usum, '(A,*(I6))', advance='no') '  ', (configs(radiatives(irad)%ini)%cfg(k), k=1, nsubshells)
        write(usum, '(A,*(I6))')               '  ', (configs(radiatives(irad)%fin)%cfg(k), k=1, nsubshells)

        ! Remove temporary files
        call execute_command('rm -vf reos.inp rcsl1.out rcsl2.out cesd1.xpn cesd2.xpn rwfn1.out rwfn2.out', &
     &                       'radiativewidths: unable to remove reos.inp rcsl1.out rcsl2.out rwfn1.out ' // &
     &                       'cesd1.xpn cesd2.xpn rwfn2.out')
      end do
      write(usum,*) '--------------------------------------------------------------------------------------', &
     &              ('------', j=1, 2*nsubshells)
      end subroutine radiativewidths


!
! radiativemultipole --- compute widths of radiative transitions for a specific multipole moment
!

      real(dblprec) function radiativemultipole(configs, radtran, multipole)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(radiativetransition), intent(in) :: radtran
      character(len=2), intent(in) :: multipole
      character(len=leninp) line, str                                                                    ! Line of input from a file
      integer i, l, levnum
      real(dblprec) radwidth, levwidth, lifetime, sumsf, sf

      ! Compute radiative transition with xreos
      open(unit=ureos, file='reos.inp', status='replace', action='write', form='formatted', access='sequential')
      ! Enter a file name for the  reos.sum  file:
      write(ureos, '(A)') 'reos.sum'
      ! Restart a previously aborted calculation ?
      write(ureos, '(A)') 'n'
      ! Enter the name of the isotope data file:
      write(ureos, '(A)') 'isodata'
      ! Enter the transition multipoles, e.g.  E1 M2 ... :
      write(ureos, '(A)') multipole
      ! Which units are to be used to enter and to print the energies of the continuum orbitals ?
      !     A       : Angstrom;
      !     eV      : electron volts;
      !     Hartree : Hartree atomic units;
      !     Hz      : Hertz;
      !     Kayser  : [cm**(-1)];
      write(ureos, '(A)') 'eV'
      ! Modify default set-up and printout of the program ?
      write(ureos, '(A)') 'y'
      ! Select individual transitions ?
      write(ureos, '(A)') 'n'
      ! Assume orthogonality between the orbital sets of the initial and final atomic states ?
      !  This feature accelerates the computations but also  neglects all relaxation effects
      !  on the transitions probabilities.
      write(ureos, '(A)') 'n'
      ! Calculate all radial integrals only once at the beginning of the computation ?
      !  This feature accelerates the computations but requires considerable more memory,
      !  in particular, if many transitions need to be calculated.
      write(ureos, '(A)') 'y'
      ! Enable the restart of the calculation if the run cannot be completed successfully ?
      !  This feature dumps all necessary information to a restart file from which the calculation can be continued.
      write(ureos, '(A)') 'n'
      ! Sort transitions in ascending order of energy ?
      write(ureos, '(A)') 'y'
      ! Read in and apply experimental energies for the calculation of transition probabilities ?
      write(ureos, '(A)') 'n'
      ! Einstein A and B coefficients are printed in SI units;
      !  use Hartree atomic units instead ?
      write(ureos, '(A)') 'n'
      ! Print all selected transitions and their energies before the computation starts (this is the default) ?
      write(ureos, '(A)') 'y'
      ! Write out the transition energies and amplitudes to an .trn file for further data processing,
      !  for instance, to adopt them to experimental transition energies ?
      write(ureos, '(A)') 'n'
      ! Fix a cutoff criterium other than '10e-8' to neglect small
      !  admixtures to the computation of transition probabilities ?
      write(ureos, '(A)') 'n'
      ! The physical speed of light in atomic units is   137.03598950000000      ;
      !  revise this value ?
      write(ureos, '(A)') 'n'
      ! The default radial grid parameters for this case are:
      !  rnt =    ....................... ;
      !  h   =    ....................... ;
      !  hp  =    ....................... ;
      !  n   =         .... ;
      !  revise these values ?
      write(ureos, '(A)') 'n'
      ! Enter the name of the initial-state CESD .xpn expansion file
      write(ureos, '(A)') 'cesd1.xpn'
      ! Enter the name of the final-state CESD .xpn expansion file
      write(ureos, '(A)') 'cesd2.xpn'
      ! Enter the name of the final-state Radial WaveFunction File:
      write(ureos, '(A)') 'rwfn1.out'
      ! Enter the name of the final-state Radial WaveFunction File:
      write(ureos, '(A)') 'rwfn2.out'
      close(ureos)
      call execute_command('$RATIP/xreos <reos.inp', 'radiativewidths: xreos returned an error')

      ! Extract radiative width
      open(unit=ureos, file='reos.sum', status='old', action='read', form='formatted', access='sequential')
      do
        read(ureos, '(A)') line
        if (trim(line) == 'Radiative lifetimes, total rates, and widths:') then
          read(ureos, '(6/)')
          radwidth = 0.0D0
!          sumsf = 0.0D0
          i = 1
          do l=1, configs(radtran%ini)%nlevels
            read(ureos, '(A)', end=10) line
            if (trim(line) == '') then
              exit
            end if
            read(line, '(I5)') levnum
            read(line, '(T30,E13.7)') lifetime
            levwidth = 1.0D0 / (lifetime * SecondTot0) * HartreeToeV
            sf = 2.0D0 * real(configs(radtran%ini)%J1(levnum), dblprec) &
     &                 / real(configs(radtran%ini)%J2(levnum), dblprec) + 1.0D0
            radwidth = radwidth + sf * levwidth
!            sumsf = sumsf + sf
            i = l
            read(ureos, '(/)', end=10)
          end do
          ! Test whether a level has zero decay width
 10       if (i < configs(radtran%ini)%nlevels) then
            write (ulog,*) '      Level with zero radiative decay width present'
          end if
          ! Sum up the statistical factor.  If a level has zero decay width it is omitted in the above sum.
          ! Thus the sum needs to be computed separately
          sumsf = 0.0D0
          do l=1, configs(radtran%ini)%nlevels
            sf = 2.0D0 * real(configs(radtran%ini)%J1(l), dblprec) &
     &                 / real(configs(radtran%ini)%J2(l), dblprec) + 1.0D0
            sumsf = sumsf + sf
          end do
          radwidth = radwidth / sumsf
          radiativemultipole = radwidth
          exit
        end if
      end do
      close(ureos)

      ! Treat files from the xreos computation
      write(line, '(A,*(A,I0))') 'fluorescence', ('_', configs(radtran%ini)%cfg(j), j = 1, nsubshells)
      write(str, '(A,*(A,I0))') '_to',           ('_', configs(radtran%fin)%cfg(j), j = 1, nsubshells)
      line = trim(line) // trim(str) // '_' // multipole // '.sum'
      call execute_command('mv -vf reos.sum fluorescence/' // trim(line), &
     &                     'radiativewidths: unable to move reos.sum to fluorescence/' // trim(line))
      call execute_command('gzip -vf9 fluorescence/' // trim(line), &
     &                     'radiativewidths: unable to compress fluorescence/' // trim(line))
      end function radiativemultipole


!
! crosssectionenergies --- generate cross sections and compute energies
!

      subroutine crosssectionenergies(configs, crosses, ncross, nxsects, maxxsectenergy)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(crosssection), allocatable, intent(out) :: crosses(:)
      integer, intent(out) :: ncross
      integer, intent(in) :: nxsects
      real(dblprec), intent(in) :: maxxsectenergy
      integer ini, fin, i, j, k, l, icrs, diffcfg(maxconf)
      real(dblprec) crossenergy, xsenbeg, xsenend, levdiff
      type(crosssection), allocatable :: tmpcross(:)

      ! Identify photoionization channels and compute threshold ionization energies
      write(ulog,'(3/,2(A/),/)') 'Photoionization thresholds:', '---------------------------'
      write(ulog,*) '   Crs. sec.   In. cfg.   Fi. cfg.      Threshold   ', &
     &              (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &              (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
      write(ulog,*) '                                           [eV]      ' &
     &           // '   ', (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      allocate(crosses(1))
      ncross = 0
      do ini=1, nconfig
 trans: do fin=ini+1, nconfig
          ! Check the difference of the configurations for potential photoionization transitions
          diffcfg = configs(ini)%cfg - configs(fin)%cfg
          l = 0 ; i = 0; j = 0
          do k=1, nsubshells
            ! Only nonzero elements matter
            if (diffcfg(k) /= 0) then
              l = l + 1
              ! More than one nonzero elements?  Then this is no photoionization
              if (l > 1) then
                cycle trans
              ! The initial state has an electron more than the final state
              else if (l == 1 .and. diffcfg(k) /= 1) then
                cycle trans
              end if
            end if
          end do
          ! If there is no nonzero element then this is no photoionization either
          if (l /= 1) then
            cycle trans
          end if

          ! Photoionizations occur if there are initial states with a higher energy than the final states
          if (.not. configs(ini)%conv .or. .not. configs(fin)%conv) then
            write (ulog, '(A,*(I3))', advance='no') ' Incomplete atomic electronic structure from ', &
     &                                        (configs(ini)%cfg(k), k=1, nsubshells)
            write (ulog, '(A,*(I3))') ' to ', (configs(fin)%cfg(k), k=1, nsubshells)
            cycle trans
          end if

          ! Threshold ionization energy of the multiplet
          if (maxval(-configs(ini)%energy) >= minval(-configs(fin)%energy)) then
            ncross = ncross + 1
            crossenergy = minval(-configs(ini)%energy) - maxval(-configs(fin)%energy)
            if (crossenergy < 0.0D0) then
              write (ulog, *) 'Overlap of initial and final state multiplets:'
              write (ulog, *) 'Initial state: [', minval(configs(ini)%energy), ' ; ', maxval(configs(ini)%energy), ']'
              write (ulog, *) 'Final state:   [', minval(configs(fin)%energy), ' ; ', maxval(configs(fin)%energy), ']'
              crossenergy = huge(0.0D0)
              ! Find the minimum level energy difference larger than zero
              do i=1, configs(ini)%nlevels
                do j=1, configs(fin)%nlevels
                  levdiff = -configs(ini)%energy(i) - (-configs(fin)%energy(j))
                  if (levdiff > 0.0D0 .and. levdiff < crossenergy) then
                    crossenergy = levdiff
                  end if
                end do
              end do
              write(ulog, *) crossenergy
            end if

            write (ulog, '(3(A,I6))', advance='no') '   ', ncross, '     ', ini, '     ', fin
            write (ulog, '(A,F15.6)', advance='no') '     ', crossenergy
            write (ulog, '(A,*(I6))', advance='no') '  ', (configs(ini)%cfg(k), k=1, nsubshells)
            write (ulog, '(A,*(I6))')               '  ', (configs(fin)%cfg(k), k=1, nsubshells)

          ! No initial state with a lower energy than all final states
          else
            write (ulog, '(A)', advance='no') 'No initial state with a lower energy than all final states:'
            write (ulog, '(A,*(I6))', advance='no') ' from ', (configs(ini)%cfg(k), k=1, nsubshells)
            write (ulog, '(A,*(I6))')               ' to '  , (configs(fin)%cfg(k), k=1, nsubshells)
            cycle trans
          end if

          ! Store photoionization cross section
          if (ncross > 1) then
            allocate(tmpcross(ncross-1))
            tmpcross(:) = crosses(1:ncross-1)
            deallocate(crosses)
            allocate(crosses(ncross))
            crosses(1:ncross-1) = tmpcross(:)
            deallocate(tmpcross)
!            call move_alloc(crosses, tmpcross)
!            allocate(crosses(ncross))
!            crosses(1:ncross-1) = tmpcross(:)
!            deallocate(tmpcross)
          end if

          crosses(ncross)%ini       = ini
          crosses(ncross)%fin       = fin
          crosses(ncross)%ionthresh = crossenergy
        end do trans
      end do

      ! Generate a linear grid close to the ionization threshold and a logarithmic grid beyond for cross sections
      do icrs=1, ncross
        allocate(crosses(icrs)%energy(nxsects))
        ! Add a small energy shift to the ionization threshold energy to make sure that it is above threshold
        ! in xphoto computations
        xsenbeg = crosses(icrs)%ionthresh + threshshift
        xsenend = crosses(icrs)%ionthresh + linenergy
        do i=1, nlingrid
          crosses(icrs)%energy(i) = xsenbeg &
     &           + (xsenend - xsenbeg) / real(nlingrid, dblprec) * real(i - 1, dblprec)
        end do
        xsenbeg = log10(xsenend)
        xsenend = log10(maxxsectenergy)
        do i=1, nxsects - nlingrid
          crosses(icrs)%energy(i+nlingrid) = 10.0D0**( xsenbeg &
     &           + (xsenend - xsenbeg) / real(nxsects - nlingrid - 1, dblprec) * real(i - 1, dblprec) )
        end do
      end do
      write(ulog,*) '-------------------------------------------------------', &
     &             ('------', j=1, 2*nsubshells)
      end subroutine crosssectionenergies


!
! crosssectionvalues --- compute values of photoionization cross sections
!

      subroutine crosssectionvalues(configs, crosses, firstcrosssection, lastcrosssection, nxsects, maxxsectenergy)
      implicit none

      type(configuration), intent(in) :: configs(:)
      type(crosssection), allocatable, intent(inout) :: crosses(:)
      integer, intent(in) :: firstcrosssection, lastcrosssection, nxsects
      real(dblprec), intent(in) :: maxxsectenergy
      logical fileexists                                                                         ! Test whether a file exists or not
      character(len=leninp) line, str                                                                    ! Line of input from a file
      integer i, j, k, J1, J2, icrs, ien
      real(dblprec) omega, omegafrac, cross, sf, sumsf

      ! Create directory for xphoto files
      inquire(file='photoionization', exist=fileexists)
      if (.not. fileexists) then
        call execute_command('mkdir -v photoionization', 'crosssectionvalues: cannot create directory photoionization')
      end if

      ! Compute cross sections of all photoionizations
      write (ulog,'(3/,2(A/))') 'Photoionization cross sections:', '-------------------------------'
      write (usum,'(3/,2(A/))') 'Photoionization cross sections:', '-------------------------------'
      do icrs=firstcrosssection, lastcrosssection
        write(ulog, '(I6,A,*(I3))', advance='no') icrs, ' Photoionization from ', &
     &                                                 (configs(crosses(icrs)%ini)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,*(I3))', advance='no') ' to ', (configs(crosses(icrs)%fin)%cfg(k), k=1, nsubshells)
        write(ulog, '(A,F15.8,A)') ' threshold energy: ', crosses(icrs)%ionthresh, ' eV'

        ! Check for acceptable number of photoionization lines
        i = configs(crosses(icrs)%ini)%nlevels * configs(crosses(icrs)%fin)%nlevels * nxsects
        if (i > maxphotolines) then
          write(ulog, '(A,I0)') '       Too many photoionization lines: ', i

          write(usum,*)
          write(usum,*) '   Crs. sec.   In. cfg.   Fi. cfg.      Threshold  ', &
     &                  (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &                  (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
          write(usum,*) '                                           [eV]        ', &
     &                  (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
          write(usum, '(3(A,I6),A,F15.6)', advance='no') '   ', icrs, '     ', crosses(icrs)%ini, '     ', &
     &                                    crosses(icrs)%fin, '    ', crosses(icrs)%ionthresh
          write(usum, '(A,*(I6))', advance='no') '  ', (configs(crosses(icrs)%ini)%cfg(k), k=1, nsubshells)
          write(usum, '(A,*(I6))')               '  ', (configs(crosses(icrs)%fin)%cfg(k), k=1, nsubshells)
          write(usum, '(/,A,I0,/)') '    Too many photoionization lines: ',i
          cycle
        end if

        ! Gather files for xphoto computation
        inquire(file='photo.sum', exist=fileexists)
        if (fileexists) then
          call execute_command('rm -vf photo.sum', 'crosssectionvalues: unable to remove photo.sum')
        end if
        write (line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(crosses(icrs)%ini)%cfg(j), j=1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/rcsl_0.out rcsl1.out', &
     &                       'crosssectionvalues: unable to copy rcsl_0.out to rcsl1.out')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn1.out', &
     &                       'crosssectionvalues: unable to copy rwfn_0.out to rwfn1.out')
        call execute_command('cp -fv ' // trim(line) // '/relci.mix relci1.mix', &
     &                       'crosssectionvalues: unable to copy relci.mix to relci1.mix')
        write (line, '(A,*(A,I0))') 'structure/rscf', ('_', configs(crosses(icrs)%fin)%cfg(j), j=1, nsubshells)
        call execute_command('cp -fv ' // trim(line) // '/rcsl_0.out rcsl2.out', &
     &                       'crosssectionvalues: unable to copy rcsl_0.out to rcsl2.out')
        call execute_command('cp -fv ' // trim(line) // '/rwfn_0.out rwfn2.out', &
     &                       'crosssectionvalues: unable to copy rwfn_0.out to rwfn2.out')
        call execute_command('cp -fv ' // trim(line) // '/relci.mix relci2.mix', &
     &                       'crosssectionvalues: unable to copy relci.mix to relci2.mix')

        ! Compute photoionization cross sections with xphoto
        open(unit=uphot, file='photo.inp', status='replace', action='write', form='formatted', access='sequential')
        ! Enter a file name for the  photo.sum  file:
        write(uphot, '(A)') 'photo.sum'
        ! Enter the name of the isotope data file:
        write(uphot, '(A)') 'isodata'
        ! Enter the transition multipoles, e.g.  E1 M2 ... :
        write(uphot, '(A)') 'E1'
        ! Which units are to be used to enter and to print the energies of the continuum orbitals ?
        !     A       : Angstrom;
        !     eV      : electron volts;
        !     Hartree : Hartree atomic units;
        !     Hz      : Hertz;
        !     Kayser  : [cm**(-1)];
        write(uphot, '(A)') 'eV'
        ! Enter the maximal energy of the photo lines  (in eV) to built-up the radial grid;
        write(uphot, '(I0)') ceiling(maxxsectenergy + 50.0D0)
        ! The photon energies can be entered either individually or by
        !  an appropriate interval and step size.
        !  The default is a list of individual energies; revise this ?
        write(uphot, '(A)') 'n'
        ! Enter another (positive) photon energy in eV enter a negative number if done.
        do i=1, nxsects
          write(uphot, *) crosses(icrs)%energy(i)
        end do
        write(uphot, '(A)') '-1.0D0'
        ! Modify default set-up and printout of the program ?
        write(uphot, '(A)') 'y'
        ! Select individual transitions ?
        write(uphot, '(A)') 'n'
        ! Include exchange interactions into the generation of the continuum waves ?
        write(uphot, '(A)') 'n'
        ! Calculate angular distribution parameters ?
        write(uphot, '(A)') 'n'
        ! Calculate spin polarization parameters ?
        write(uphot, '(A)') 'n'
        ! Sort all photoionization lines in ascending order of energy ?
        write(uphot, '(A)') 'n'
        ! Read in and apply experimental energies for the calculation of Auger rates and other properties ?
        write(uphot, '(A)') 'n'
        ! Photoionization cross sections are printed in SI units; use Hartree atomic units instead ?
        write(uphot, '(A)') 'n'
        ! Print all selected photoionization lines before the computation starts (this is the default) ?
        write(uphot, '(A)') 'y'
        ! Print the CSF scheme each time a new one has been built ?
        write(uphot, '(A)') 'n'
        ! Print the results for each individual line immediatly after its computation ?
        write(uphot, '(A)') 'n'
        ! Print the final results to a (.chn) channel amplitude file ?
        write(uphot, '(A)') 'n'
        ! Enter an (overall) shift for the atomic transition energies which applies to all transitions (ineV):
        !  Use  0.  or   <cr>  if no shift need to be applied.
        write(uphot, '(A)') '0'
        ! Enter a maximal (-)kappa symmetry up to which continuum spinors are taken into account ?
        !  2 (up to p-waves), 4(f), 6(i), 8(k), ...; 0 or <cr> to include all possible waves.
        write(uphot, '(A)') '0'
        ! The physical speed of light in atomic units is   137.03598950000000      ;
        !  revise this value ?
        write(uphot, '(A)') 'n'
        ! The default radial grid parameters for this case are:
        !  rnt =    ....................... ;
        !  h   =    ....................... ;
        !  hp  =    ....................... ;
        !  n   =         .... ;
        !  revise these values ?
        write(uphot, '(A)') 'n'
        ! Enter the name of the initial-state GRASP92 configuration symmetry list file:
        write(uphot, '(A)') 'rcsl1.out'
        ! Enter the name of the final-state GRASP92 configuration symmetry list file:
        write(uphot, '(A)') 'rcsl2.out'
        ! Enter the name of the initial-state GRASP92 mixing coefficient file:
        write(uphot, '(A)') 'relci1.mix'
        ! Enter the name of the final-state GRASP92 mixing coefficient file:
        write(uphot, '(A)') 'relci2.mix'
        ! Enter the name of the initial-state Radial WaveFunction File:
        write(uphot, '(A)') 'rwfn1.out'
        ! Enter the name of the final-state Radial WaveFunction File:
        write(uphot, '(A)') 'rwfn2.out'
        close(uphot)
        call execute_command('$RATIP/xphoto <photo.inp', 'crosssectionvalues: xphoto returned an error')

        ! Extract photoionization cross section
        allocate(crosses(icrs)%cross(nxsects))
        crosses(icrs)%cross = 0.0D0
        open(unit=uphot, file='photo.sum', status='old', action='read', form='formatted', access='sequential')
        do
          read(uphot, '(A)') line
          if (trim(line) == ' Individual and total photoionization cross sections :') then
            read(uphot, '(5/)')
            do
              read(uphot, '(A)') line
              if (trim(line) == ' -------------------------------------------------------------------------' // &
     &                          '-------------------------------------------------------') then
                exit
              end if
              ! Read initial state total angular momentum
              if (line(16:16) == '/') then
                read(line, '(T14,I2,T17,I2)') J1, J2
              else
                read(line, '(T14,I5)') J1
                J2 = 1
              end if
              sf = 2.0D0 * real(J1, dblprec) / real(J2, dblprec) + 1.0D0
              ! Read photon energy and cross section in Babushkin gauge
              read(line, '(T36,E11.5,T70,E9.3)') omega, cross
              cross = cross * cross_section_MegaBarntocm2
              ! Read only the fractional part of the photon energy
              read(line, '(T36,F7.5)') omegafrac
              ! Search for photon energy
              ien = 1
              do
                ! Output energies in photo.sum have an accuracy of 6 digits
                if (abs(omega - crosses(icrs)%energy(ien)) < 1.0D-5 * omega / omegafrac) then
                  crosses(icrs)%cross(ien) = crosses(icrs)%cross(ien) + sf * cross
                  exit
                end if
                ien = ien + 1
                if (ien > nxsects) then
                  call error_stop('crosssectionvalues: not matching photon energy')
                end if
              end do
            end do

            ! Calculate the sum of the probabilistic factor
            sumsf = 0.0D0
            do i=1, configs(crosses(icrs)%ini)%nlevels
              sf = 2.0D0 * real(configs(crosses(icrs)%ini)%J1(i), dblprec) &
     &                   / real(configs(crosses(icrs)%ini)%J2(i), dblprec) + 1.0D0
              sumsf = sumsf + sf
            end do
            crosses(icrs)%cross = crosses(icrs)%cross / sumsf

            ! Print photoionization cross section
            write(usum,*)
            write(usum,*) '   Crs. sec.   In. cfg.   Fi. cfg.      Threshold  ', &
     &                    (' ', j=1, (6 * nsubshells - 8 + 4) / 2), 'In. Sub.', &
     &                    (' ', j=1,  6 * nsubshells - 8 + 4 - 4), '  Fi. Sub.'
            write(usum,*) '                                           [eV]        ', &
     &                    (subshellnames(j) // '    ', j=1, nsubshells), '  ', (subshellnames(j) // '    ', j=1, nsubshells)
            write(usum, '(3(A,I6),A,F15.6)', advance='no') '   ', icrs, '     ', crosses(icrs)%ini, '     ', &
     &                                      crosses(icrs)%fin, '    ', crosses(icrs)%ionthresh
            write(usum, '(A,*(I6))', advance='no') '  ', (configs(crosses(icrs)%ini)%cfg(k), k=1, nsubshells)
            write(usum, '(A,*(I6))')               '  ', (configs(crosses(icrs)%fin)%cfg(k), k=1, nsubshells)
            write(usum, '(/A)') '            Energy [eV]       Cross section [cm**2]'
            do j=1, nxsects
              write(usum, '(A,F15.6,A,E13.6)') '        ', crosses(icrs)%energy(j), '          ', crosses(icrs)%cross(j)
            end do
            write(usum,*)
            exit
          end if
        end do
        close(uphot)

        ! Treat files from the xphoto computation
        write(line, '(A,*(A,I0))') 'photo', ('_', configs(crosses(icrs)%ini)%cfg(j), j = 1, nsubshells)
        write(str, '(A,*(A,I0))') '_to',    ('_', configs(crosses(icrs)%fin)%cfg(j), j = 1, nsubshells)
        line = trim(line) // trim(str) // '.sum'
        call execute_command('mv -vf photo.sum photoionization/' // trim(line), &
     &                       'crosssectionvalues: unable to move auger.sum to photoionization/' // trim(line))
        call execute_command('gzip -vf9 photoionization/' // trim(line), &
     &                       'crosssectionvalues: unable to compress photoionization/' // trim(line))
        call execute_command('rm -vf photo.inp rcsl1.out rcsl2.out rwfn1.out rwfn2.out relci1.mix relci2.mix', &
     &                       'crosssectionvalues: unable to remove photo.inp rcsl1.out rcsl2.out rwfn1.out ' // &
     &                       'rwfn2.out relci1.mix relci2.mix')
      end do
      end subroutine crosssectionvalues
      end program caesr