!
! zblanc --- band Lanczos for complex Hermitian matrices
! Thomas Sommerfeld based on a real symmetric code by Hans-Dieter Meyer
! Planet PKS version by Christian Buth
! 26 July 2004
!
!
!     ----------------------------------------------------------------
!
!     BAND LANCZOS for complex symmetric matrices
!
!     large parts of this routine are taken from the block lanczos
!     program by H-D. Meyer and U. V. Riss
!
!     a matrix A is transformed to band-diagonal form T
!     A is only refered to via the subroutine zmatblk, which
!     has to perform the matrix times vector-block operation
!
!     ndim     dimension of the matrix A
!     np       block-size for the iteration --> band width of T
!     niter    no. of iterations (matrix*vector operations)
!     iparam   array of additional parameters
!     iparam(1)  jstart     1st iteration  (can be > 1 in restarts)
!     iparam(2)  jend       index of last iteration (output)
!     iparam(3)  neigenv    abs of no. of eigenvectors to be computed
!                neigenv =  0  : normal Lanczos
!                neigenv = -1  : preliminary run for eigenvectors
!                neigenv >  0  : eigenvector run
!     iparam(4)  restart
!     iparam(5)  wrtrsf    write a restart file
!     iparam(6)  wrttmat   write T to a file
!     iparam(7)  ncartsv   no. of cartesian start-vectors
!                if the start-vectors are cartesian this is used
!                to normalize the Lanczos-vectors on the start-block
!
!     t        contains on output the computed band matrix
!     p        contains 3*np vectors needed in the iteration
!              on input p(:,i,2) is used as the i-th start-vector
!
!     z and ev are only refered to if iparam(3) > 0
!     z        contains some eigenvectors of T from a previous run
!     ev       contains on output the corresponding eigenvectors of A
!
!     icartsv  contains the indices of the cartesian start-vectors
!
!     external zmatblk(ndim, np, pold, pnew, work)
!     is the mat times vector routine which computes the product
!     between the np vectors pold and the matrix and stores the
!     np resulting vectors in pnew
!     work contains space for zmatvec and is not referenced in the Lanczos
!

      subroutine zblanc(ndim, np, niter, neigenv, ncartsv, iparam, t,
     &          p, z, ev, icartsv, zmatblk)
      use lanczos, zblancdbl => zblanc
      implicit none

      integer, intent(in) :: ndim, np, niter, neigenv,
     &          iparam(10)
      integer, intent(inout) :: ncartsv, icartsv(ncartsv)
      complex(dblprec), intent(inout) :: p(ndim,np,3), ev(ndim,neigenv),
     &          z(niter,neigenv)
      complex(dblprec), intent(out) :: t(np+1,niter)
      external zmatblk

      integer jstart, jend
      integer restart, wrtrsf, wrttmat
      integer np1, np3

      complex(dblprec) y
      integer i, j, k, nrun, ndone
      integer jj, ii, jn, in, jo, io, jjc, iic, im, ip, jp
      integer jiter, jko
      logical da

      complex(dblprec) zdotu
      external zdotu, zaxpy, zscal

! CB #include "iosetting.inc"

!     -------------
!     Preliminaries
!     -------------
      write (6,100)
 100  format (//,2X,50('-')/,2X,'Band-Lanczos (complex symmetric)'/)
      jstart  = iparam(1)
      jend    = iparam(2)
      restart = iparam(4)
      wrtrsf  = iparam(5)
      wrttmat = iparam(6)
      np1 = np + 1
      np3 = 3 * np
      jstart = 1
!     --------------------
!     Read the restartfile
!     --------------------
      if (restart .ne. 0) then
         write (6,'(4x,A)') 'Restart from file'
         inquire(file='RESTART', exist=da)
         if (.not. da) then
            write(6,*) 'RESTART File is missing.'
            stop
         end if
         open (ifile,file='restart',access='sequential',
     &        form='unformatted')
         read(ifile) jend, j
         if (j .ne. ndim) then
            write(6,*) 'Inconsistancy:'
            write(6,*) 'Ndim(Inputfiles)  = ', ndim
            write(6,*) 'Ndim(Restartfile) = ', j
            stop
         end if
         write(6,'(6x,A,I5)') 'Iterations done so far  ', jend
         write(6,'(6x,A,I5)') 'Iterations to be done   ', niter-jend
         if (jend .ge. niter) then
            stop 'Nothing to do'
         end if
         read(ifile) ((t(i,j), i=1, np1),  j=1, jend)
         read(ifile) ((p(i,j,1), i=1, ndim), j=1, np),
     &            ((p(i,j,2), i=1, ndim), j=1, np)
         read(ifile) ncartsv
         read(ifile) (icartsv(i), i=1, ncartsv)
         close (ifile)
         jstart = jend + 1
      end if
!     ----------------------------
!     Print Iterations and runtype
!     ----------------------------
      write (6,'(4x,A,I5,A,I5)') 'Iterations from ',jstart,' to ',niter
      if (neigenv .gt. 0) then
         write(6,'(4X,A,I4,A)') 'This is a second run computing ',
     &        neigenv,' eigenvectors'
      end if
      write(6,*)
!     ---------------------------
!     Prepare for eigenvector run
!     ---------------------------
      if (neigenv .gt. 0) then
!     --------------------------------------
!     Read the ZMAT file to z(niter,neigenv)
!     this file is created by banddia after
!     a first lanczos requiring eigenvectors
!     --------------------------------------
        inquire(file='ZMAT', exist=da)
         if (.not. da) then
            write(6,*) '**** The File ZMAT does not exist.'
            stop
         end if
         open (ifile,file='ZMAT',access='sequential',
     &        form='unformatted')
         read(ifile) i, j
         if (i .ne. niter .or. j .ne. neigenv) then
            write(6,*) '***************   WARNING  *****************'
            write(6,*) ' ZMAT file _may_ be corrupted.'
            write(6,*) ' niter, neigenv (Args) = ', niter, neigenv
            write(6,*) ' zdim,  neigenv (File) = ', i, j
            write(6,*) ' (niter must be too small in a restart set up)'
            write(6,*)
            if (j .lt. neigenv .or. i .lt. niter) then
               stop
            end if
         end if
         do j = 1, neigenv
            read(ifile) (z(i,j), i = 1, niter)
         end do
         close (ifile)
!     -------------------------------
!     initialize or read eigenvectors
!     -------------------------------
         if (restart .eq. 0) then
            do i = 1, neigenv
               do j = 1, ndim
                  ev(j,i) = (0.0d0, 0.0d0)
               end do
            end do
            do j = 1, np
               jiter = jstart - 1 + j
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i) = ev(jko,i)+p(jko,j,2)*z(jiter,i)
                  end do
               end do
            end do

         else
!     ------------------------------------------------------
!     restart = 1
!     read eigenvectors from the EIGENV file of the last run
!     ------------------------------------------------------
            inquire(file='EIGENV', exist=da)
            if (.not. da) then
               write(6,*) 'EIGENV File is missing.'
               stop
            end if
            open (ifile,file='eigenv',access='sequential',form
     &           ='unformatted')
            read(ifile) i, j
            if (i .ne. neigenv .or. j .ne. ndim) then
               write(6,*) 'EIGENV file may be corrupted.'
               write(6,*) 'ndim, neigenv (Args) = ', ndim, neigenv
               write(6,*) 'ndim, neigenv (File) = ', j, i
               stop
            end if
            read(ifile) i
            if (i .ne. 2) then
               write(6,*) 'itype = ',i,' ?????'
               stop
            end if
            do j = 1, neigenv
               read(ifile) (ev(i,j), i = 1, ndim)
            end do
            close(ifile)
!     ------------------------------------------
!     contributions of the restart vectors which
!     have NOT been added in the previous run!
!     ------------------------------------------
            do j = 1, np
               jiter = jstart - 1 + j
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i) = ev(jko,i) + p(jko,j,2) * z(jiter,i)
                  end do
               end do
            end do
         end if
      end if
!     ==============================================
!     end of eigenvector preparations, now start the
!
!     Main Loop of the Lanczos Iteration
!     ----------------------------------
      nrun = niter - jstart + 1
      ndone = 10

      do 1  jj = jstart, niter


         call indexl(jj-jstart+np, np, jn, in )
!
!     when neccessary do the next matrix times block operation
!
         if (mod(jj-jstart,np) .eq. 0) then
            io = mod(in+1, 3) + 1
            call zmatblk(ndim, np, p(1,1,io), p(1,1,in))
         end if
!
!     Band-Lanczos algorithm as implemented by H-D. Meyer
!
         do II = MAX0(JJ-NP,1), JJ-1
            call indexl(ii-jstart+np3, np, jo, io)
            JJC = JJ - II + 1
!            do K = 1, NDIM
!               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*T(JJC,II)
!            end do
            call zaxpy(ndim,-t(jjc,ii),p(1,jo,io),1,p(1,jn,in),1)
         end do

         do II = JJ, JJ + NP - 1
            CALL INDEXL(II-JSTART, NP, JO, IO )
!            Y = (0.0d0, 0.0d0)
!            do K = 1, NDIM
!               Y = Y + P(k,JO,IO) * P(k,JN,IN)
!            end do
            y = zdotu(ndim,p(1,jo,io),1,p(1,jn,in),1)
!            do K = 1, NDIM
!               P(k,JN,IN) = P(k,JN,IN) - P(k,JO,IO)*Y
!            end do
            call zaxpy(ndim,-y,p(1,jo,io),1,p(1,jn,in),1)
            IIC = II - JJ + 1
            T(IIC,JJ) = Y
         end do
!     -------------------------------------------------------------
!     orthogonalization with respect to the cartesian start-vectors
!     Vorsicht bei restart; icartsv muss definiert sein
!     -------------------------------------------------------------
         if (ncartsv .gt. 0) then
            do k = 1, np
               p(icartsv(k),jn,in) = (0.0d0, 0.0d0)
            end do
         end if
!     --------------------
!     normalize new vector
!     --------------------
!         Y = (0.0d0, 0.0d0)
!         do K = 1, NDIM
!            Y = Y + P(k,JN,IN)**2
!         end do
         y = zdotu(ndim,p(1,jn,in),1,p(1,jn,in),1)
         y = sqrt(Y)
         T(NP1,JJ) = Y
         y = 1.0D0 / y
!         do K = 1, NDIM
!            P(k,JN,IN) = P(k,JN,IN) * y
!         end do
         call zscal(ndim,y,p(1,jn,in),1)
!     -------------------------------------------------------
!     sum contributions of this iteration to the eigenvectors
!     -------------------------------------------------------
         if (neigenv .gt. 0) then
            jiter = np + jj
!
!     in the eigenvector calculation the very last np iterations
!     do not contribute to the eigenvectors;
!     but in case of a later restart these iterations have to be done
!
            if (jiter .le. niter) then
               do i = 1, neigenv
                  do jko = 1, ndim
                     ev(jko,i)=ev(jko,i)+z(jiter,i)*p(jko,JN,IN)
                  end do
               end do
            end if
         end if
!
!     check how much has been done
!
         if ((jj-jstart+1)*100 / nrun .ge. ndone) then
            write(6,'(6X,I3,A)') ndone, '%'
! CB
!            call flush(6)
! END CB
            ndone = ndone + 10
         end if

 1    CONTINUE
!     ---------------------------------------------------------
!     End of the Block-Lanczos iteration
!     jend was used when the main-loop was left because of
!     a timelimit (IMB 30/90 version); this has been thrown out
!     ---------------------------------------------------------
      write(6,999)
 999  FORMAT(/,2X,'Lanczos finished',/,2X,50('-'))
      jend = niter
!     ---------------------------------------------------
!     write the bandmatrix T(np1, niter) to the file TMAT
!     for convergence and other checks
!     ---------------------------------------------------
      if (wrttmat .eq. 1) then
         open(unit=ifile,file='TMAT',access='SEQUENTIAL',
     &        form='UNFORMATTED')
         write (ifile) jend, np, niter
         write (ifile) t
         close (ifile)
      end if
!     -----------------------------------------------
!     write the restart file
!     this part was written by Dieter;
!     i do not really understand it, however it works
!     -----------------------------------------------
      if (wrtrsf .eq. 1) then
         open (ifile,file='RESTART',access='SEQUENTIAL',
     &        form='UNFORMATTED')
         CALL INDEXL(JEND-JSTART+NP, NP, JP, IP )
         IO  = MOD(IP+1,3) + 1
         IM  = MOD(IP  ,3) + 1
         write(ifile) jend, ndim
         write(ifile) ((T(I,J), I=1,NP1) ,J=1,JEND)
         write(ifile) ((P(i,j,IM), i=1,NDIM), j=JP+1,NP),
     &             ((P(i,j,IO), i=1,NDIM), j=1,   NP),
     &             ((P(i,j,IP), i=1,NDIM), j=1,   JP)
         write(ifile) ncartsv
         write(ifile) (icartsv(i), i=1, ncartsv)
         close (ifile)
      end if


      contains

!
!     one of Dieter's more serious crimes:
!
!     THE SUBROUTINE INDEXL COMPUTES THE CURREND INDEX OF THE LANCZOS-
!                           VECTOR CONTAINED IN THE ARRAY P.
!
!     AH1:LANCZ.INDEXL
!     THIS ROUTINE COMPUTES THE INDICES J,I OF THE JJ-TH LANCZOS-VECTOR
!

      subroutine indexl( jj, np, j, i )
      implicit none

      integer, intent(in) :: jj, np
      integer, intent(out) :: j, i
      integer kk

      KK = JJ/NP
      I  = MOD( KK+1, 3 ) + 1
      J  = JJ + 1 - KK*NP
      end subroutine indexl
      end subroutine zblanc