!
! SESpath --- Complex path for smooth exterior complex scaling (SES)
! Christian Buth
! 23 May 2006
!
! Version V1.00: 1 Jun 2006 (Numerically stable evaluation of the
!       path in the complex plane)
! Version V1.10: 30 Aug 2006 (Conventional complex scaling from
!       the origin)
! Version V1.20: 14 Nov 2008 (Double scaling for conventional complex
!       scaling eliminated)
!


      module SESpath
      use fortrantl
      implicit none

!#define fltprec quadprec
#define fltprec dblprec

      logical originCS                          ! Do complex scaling from r=0 on

      real(fltprec), parameter :: replacethresh = real(1.0D-8, fltprec), &      ! Cutoff for
                        ! replacing analytical expressions for the complex
                        ! path function
     &          expcutoff = 280.0D0             ! For stable numerical
                        ! evaluation it is necessary to restrict the exponent
                        ! of the exp function in the calculation of the paths

      real(fltprec) sestheta,  &                ! Imke: Complex phase; Nimrod theta = 0.75
     &          seslambda,     &                ! Smoothness of the path, Nimrod lambda = 5
     &          sesr0,         &                ! Start distance of the exterior scaling
     &          sesrlow,       &                ! For r < sesrlow,  the complex path is rho = r
     &          sesrhigh                        ! For r > sesrhigh, the complex path is rho = r + eitmo * (r - r0)
      complex(fltprec), private :: eitmo        ! e^(i*theta) - 1
      complex(fltprec) eit                      ! e^(i*theta)


      contains

!
! initSESpath --- initialize the SESpath module.  Traverse the complex
!       scaling path F(r) to find the region where a precise numerical
!       evaluation is required
!

      subroutine initSESpath(theta, lambda, r0)
      implicit none

      real(dblprec), intent(in) :: theta, lambda, r0
      real(fltprec) rstep
      complex(fltprec) rho

      sestheta  = real(theta , fltprec)
      seslambda = real(lambda, fltprec)
      sesr0     = real(r0    , fltprec)
      eit       = cmplx(0.0D0, 1.0D0, fltprec) * cmplx(theta, 0.0D0, fltprec)
      eit       = cmplx(exp(cmplx(eit, kind=fltprec)), kind=fltprec) 
!      eitmo     = cmplx(0.0D0, 1.0D0, fltprec) * cmplx(theta, 0.0D0, fltprec)
!!       eitmo     = exp(eitmo) - cmplx(1.0D0, 0.0D0, fltprec)
!!Bug in g95! delete line below and replace with line above
!      eitmo     = cmplx(exp(cmplx(eitmo, kind=dblprec)), kind=fltprec)  &
!     &            - cmplx(1.0D0, 0.0D0, fltprec)
      eitmo     = eit - cmplx(1.0D0, 0.0D0, fltprec)
!      write (*,*) '           My eit =', eit, ' eitmo=', eitmo

      ! Conventional complex scaling from the origin
      if (originCS) then
        write (*,*) 'Conventional complex scaling from the origin'
        write (*,*) '    Scaling angle theta =', theta
        return
      end if
     
      ! Smooth exterior complex scaling
      write (*,*) 'Parameters for the smooth exterior scaling path'
      write (*,*) '    scaling angle theta =', theta
      write (*,*) '    path smoothness lambda =', lambda
      write (*,*) '    begin scaling r0 =', r0

      rstep = seslambda * real(2.0D0 * 0.002D0, fltprec)
      write (*,*) 'Traversing the complex path with rstep = ', rstep, &
     &            ' for replacethresh = ', replacethresh
      if (abs(r0) < 2.0D0 * rstep .or. r0 < 0.0D0) then
        stop 'initSESpath: abs(r0) < 10.0D0 * rstep .or. r0 < 0.0D0 !'
      end if

      sesrlow = sesr0
      do
        ! The path for 0 <= r << r0
        rho = sesrlow
        if (abs(rho - evalFc(real(sesrlow, dblprec))) < replacethresh) then
          exit
        end if
        sesrlow = sesrlow - rstep
        if (sesrlow <= 0.0D0) then
          sesrlow = 0.0D0
          exit
           stop 'initSESpath: sesrlow <= 0.0D0 !'
        end if
      end do
      write (*,*) '    lower end sesrlow  = ', sesrlow

      sesrhigh = sesr0
      do
        ! The path for r >> r0
        rho = sesrhigh + eitmo * (sesrhigh - sesr0)
        if (abs(rho - evalFc(real(sesrhigh, dblprec))) < replacethresh) then
          exit
        end if
        sesrhigh = sesrhigh + rstep
        if (abs(2.0D0*lambda * (sesrhigh+r0)) >= expcutoff) then
          write (*,*) 'abs(2.0D0*lambda * (sesrhigh+r0)) =',  &
     &                 abs(2.0D0*lambda * (sesrhigh+r0)),     &
     &                 'expcutoff =', expcutoff
          stop 'initSESpath: abs(2.0D0*lambda * (sesrhigh+r0)) >= expcutoff !'
        end if
      end do

      write (*,*) '    upper end sesrhigh = ', sesrhigh
      end subroutine initSESpath


!
! exitSESpath --- clean up module
!

      subroutine exitSESpath
      implicit none

      end subroutine exitSESpath


!
! evalFc --- evaluate the analytic expression of the complex
!       scaling path F(r)
!

      complex(dblprec) function evalFc(rarg)
      implicit none

      real(dblprec), intent(in) :: rarg
      real(fltprec) r

      r = real(rarg, fltprec)

      ! Path only defined for r > 0
      if (r < 0.0D0) then
        stop 'evalFc: r < 0.0D0 !'
      end if

      ! Required for the numerical stability of path evaluation
      if (abs(2.0D0*seslambda * (r-sesr0)) >= expcutoff .or.  &
     &    abs(2.0D0*seslambda * (r+sesr0)) >= expcutoff) then
        stop 'evalFc: abs(2.0D0*seslambda * (r-sesr0)) >= expcutoff .or. abs(2.0D0*seslambda * (r+sesr0)) >= expcutoff'
      end if

      ! The path in the complex plane
      evalFc = r + eitmo * (r + sesr0 + 0.5D0 / seslambda          &
     &         * log(  (1.0D0 + exp(2.0D0*seslambda * (r-sesr0)))   &
     &               / (1.0D0 + exp(2.0D0*seslambda * (r+sesr0)))))
      end function evalFc


!
! pathFc --- the complex scaling path F(r)
!

      complex(dblprec) function pathFc(rarg)
      implicit none

      real(dblprec), intent(in) :: rarg
      real(fltprec) r

      r = real(rarg, fltprec)

      ! Path only defined for r > 0
      if (r < 0.0D0) then
        stop 'pathFc: r < 0.0D0 !'
      end if

      ! Conventional complex scaling from the origin
      if (originCS) then
        pathFc = ((1.0D0, 0.0D0) + eitmo) * cmplx(r, kind=dblprec)
      ! Smooth exterior complex scaling
      else
        if (sesrlow == 0.0D0 .and. sesrhigh == 0.0D0) then
          stop 'pathFc: call initSESpath first!'
        end if
        
        ! The path in the complex plane
        if (r < sesrlow) then
          pathFc = r
        else if (r > sesrhigh) then
          pathFc = r + eitmo * (r - sesr0)
        else
          pathFc = evalFc(real(r, dblprec))
        end if
      end if
      end function pathFc


!
! pathf --- derivative of complex scaling path f(r) = F'(r)
!

      complex(dblprec) function pathf(rarg)
      implicit none

      real(dblprec), intent(in) :: rarg
      real(fltprec) r

      r = real(rarg, fltprec)

      ! Path only defined for r > 0
      if (r < 0.0D0) then
        stop 'pathf: r < 0.0D0 !'
      end if

      ! Conventional complex scaling from the origin
      if (originCS) then
        pathf = (1.0D0, 0.0D0) + eitmo
      ! Smooth exterior complex scaling
      else
        if (sesrlow == 0.0D0 .and. sesrhigh == 0.0D0) then
          stop 'pathf: call initSESpath first!'
        end if

        if (abs(2.0D0*seslambda * (r+sesr0)) < expcutoff .and. &
     &      abs(2.0D0*seslambda * (r-sesr0)) < expcutoff) then
          pathf = (1.0D0, 0.0D0) + eitmo * (1.0D0 + 1.0D0 / (1.0D0 + exp(2.0D0*seslambda * (r+sesr0)))  &
     &                                   - 1.0D0 / (1.0D0 + exp(2.0D0*seslambda * (r-sesr0))))
        else if (abs(2.0D0*seslambda * (r+sesr0)) < expcutoff .and. &
     &      abs(2.0D0*seslambda * (r-sesr0)) >= expcutoff) then
          pathf = (1.0D0, 0.0D0) + eitmo * (1.0D0 + 1.0D0 / (1.0D0 + exp(2.0D0*seslambda * (r+sesr0))))
        else if (abs(2.0D0*seslambda * (r+sesr0)) >= expcutoff .and. &
     &      abs(2.0D0*seslambda * (r-sesr0)) < expcutoff) then
          pathf = (1.0D0, 0.0D0) + eitmo * (1.0D0 - 1.0D0 / (1.0D0 + exp(2.0D0*seslambda * (r-sesr0))))
        else
          pathf = (1.0D0, 0.0D0) + eitmo
        end if
      end if
      end function pathf


!
! pathfp --- second derivative of complex scaling path f'(r) = F''(r)
!

      complex(dblprec) function pathfp(rarg)
      implicit none

      real(dblprec), intent(in) :: rarg
      real(fltprec) r

      r = real(rarg, fltprec)

      ! Path only defined for r > 0
      if (r < 0.0D0) then
        stop 'pathfp: r < 0.0D0 !'
      end if

      ! Conventional complex scaling from the origin
      if (originCS) then
        pathfp = (0.0D0, 0.0D0)
      ! Smooth exterior complex scaling
      else
        pathfp = (0.0D0, 0.0D0)

        if (sesrlow == 0.0D0 .and. sesrhigh == 0.0D0) then
          stop 'pathfp: call initSESpath first!'
        end if
        if (r < sesrlow .or.r > sesrhigh) then
          return
        end if

        if (abs(2.0D0*seslambda * (r+sesr0)) < expcutoff .and. &
     &      abs(2.0D0*seslambda * (r-sesr0)) < expcutoff) then
          pathfp = 2.0D0 * seslambda * eitmo * (                       &
     &                   - exp(2.0D0*seslambda * (r+sesr0))               &
     &                   / (1.0D0 + exp(2.0D0*seslambda * (r+sesr0)))**2  &
     &                   + exp(2.0D0*seslambda * (r-sesr0))               &
     &                   / (1.0D0 + exp(2.0D0*seslambda * (r-sesr0)))**2)
        else
          pathfp = (0.0D0, 0.0D0)
        end if
      end if
      end function pathfp


!
! pathfpp --- third derivative of complex scaling path f''(r) = F'''(r)
!

      complex(dblprec) function pathfpp(rarg)
      implicit none

      real(dblprec), intent(in) :: rarg
      real(fltprec) r

      r = real(rarg, fltprec)

      ! Path only defined for r > 0
      if (r < 0.0D0) then
        stop 'pathfpp: r < 0.0D0 !'
      end if

      ! Conventional complex scaling from the origin
      if (originCS) then
        pathfpp = (0.0D0, 0.0D0)
      ! Smooth exterior complex scaling
      else
        pathfpp = (0.0D0, 0.0D0)

        if (sesrlow == 0.0D0 .and. sesrhigh == 0.0D0) then
          stop 'pathfpp: call initSESpath first!'
        end if
        if (r < sesrlow .or. r > sesrhigh) then
          return
        end if

        if (abs(2.0D0 * seslambda * (r+sesr0)) < expcutoff) then
          pathfpp =         - 1.0D0 / (exp(-seslambda * (r+sesr0))                     &
     &                               + exp( seslambda * (r+sesr0)))**2
        end if
        if (abs(2.0D0 * seslambda * (r-sesr0)) < expcutoff) then
          pathfpp = pathfpp + 1.0D0 / (exp(-seslambda * (r-sesr0))                     &
     &                               + exp( seslambda * (r-sesr0)))**2
        end if
        if (abs(4.0D0*seslambda * (r+sesr0)) < expcutoff) then
          pathfpp = pathfpp + 2.0D0 / (exp(-4.0D0/3.0D0*seslambda * (r+sesr0)) &
     &                               + exp( 2.0D0/3.0D0*seslambda * (r+sesr0)))**3
        end if
        if (abs(4.0D0*seslambda * (r-sesr0)) < expcutoff) then
          pathfpp = pathfpp - 2.0D0 / (exp(-4.0D0/3.0D0*seslambda * (r-sesr0)) &
     &                               + exp( 2.0D0/3.0D0*seslambda * (r-sesr0)))**3
        end if

        pathfpp = 4.0D0 * seslambda**2 * eitmo * pathfpp
      end if
      end function pathfpp
      end module SESpath
