!    anglib.f90: angular momentum coupling coefficients in Fortran 90
!    Copyright (C) 1998  Paul Stevenson
!


module anglib
use fortrantl
!
! Library of angular momentum coupling coefficient routines in fortran 90
! Paul Stevenson, p.stevenson@surrey.ac.uk
! Modified by Robin Santra on 10/10/02 from single to double precision.
contains

  function cleb(j1,m1,j2,m2,j,m)
    implicit none
    ! calculate a clebsch-gordan coefficient < j1/2 m1/2 j2/2 m2/2 | j/2 m/2 >
    ! arguments are integer and twice the true value.

    real(dblprec) cleb,factor,sum
    integer j1,m1,j2,m2,j,m,par,z,zmin,zmax

    ! some checks for validity (let's just return zero for bogus arguments)
    if (2*(j1/2)-int(2*(j1/2.0)) /= 2*(abs(m1)/2)-int(2*(abs(m1)/2.0)) .or. &
         2*(j2/2)-int(2*(j2/2.0)) /= 2*(abs(m2)/2)-int(2*(abs(m2)/2.0)) .or. &
         2*(j/2)-int(2*(j/2.0)) /= 2*(abs(m)/2)-int(2*(abs(m)/2.0)) .or. &
         j1<0 .or. j2<0 .or. j<0 .or. abs(m1)>j1 .or. abs(m2)>j2 .or.&
         abs(m)>j .or. j1+j2<j .or. abs(j1-j2)>j .or. m1+m2/=m) then
       cleb= 0.0d0
    else

       factor = 0.0d0
       factor = binom(j1,(j1+j2-j)/2) / binom((j1+j2+j+2)/2,(j1+j2-j)/2)
       factor = factor * binom(j2,(j1+j2-j)/2) / binom(j1,(j1-m1)/2)
       factor = factor / binom(j2,(j2-m2)/2) / binom(j,(j-m)/2)
       factor = dsqrt(factor)

       zmin = max(0,j2+(j1-m1)/2-(j1+j2+j)/2,j1+(j2+m2)/2-(j1+j2+j)/2)
       zmax = min((j1+j2-j)/2,(j1-m1)/2,(j2+m2)/2)

       sum=0.0d0
       do z = zmin,zmax
          par=1
          if(2*(z/2)-int(2*(z/2.0)) /= 0) par=-1
          sum=sum+par*binom((j1+j2-j)/2,z)*binom((j1-j2+j)/2,(j1-m1)/2-z)*&
               binom((-j1+j2+j)/2,(j2+m2)/2-z)
       end do

       cleb = factor*sum
    end if
  end function cleb


  function sixj(a,b,c,d,e,f)
    use fortrantl
    implicit none
    integer, intent(in) :: a,b,c,d,e,f
    real(dblprec) :: sixj
    integer :: phase, nlo, nhi, n
    real(dblprec) :: outfactors, sum, sumterm
    ! calculates a Wigner 6-j symbol. Argument a-f are integer and are
    ! twice the true value of the 6-j's arguments, in the form
    ! { a b c }
    ! { d e f }

    ! First check for consistency of arguments:
    sixj=0.0d0
    if(mod(a+b,2)/=mod(c,2)) return
    if(mod(c+d,2)/=mod(e,2)) return
    if(mod(a+e,2)/=mod(f,2)) return
    if(mod(b+d,2)/=mod(f,2)) return
    if(abs(a-b)>c .or. a+b<c) return
    if(abs(c-d)>e .or. c+d<e) return
    if(abs(a-e)>f .or. a+e<f) return
    if(abs(b-d)>f .or. b+d<f) return

    phase=(-1)**((a+c+d+f)/2)

    outfactors = angdelta(a,b,c)/angdelta(a,e,f)
    outfactors = outfactors * angdelta(b,d,f)/angdelta(c,d,e)

    nlo = 0
    nhi = min( (a-b+d+e)/2, (c+e+f-b)/2, (a+c+d+f)/2+1, (a-b+c)/2, &
         & (d+f-b)/2, (a+e+f)/2+1, (c+d+e)/2+1)

    sum=0.0d0
    do n=nlo,nhi
       sumterm = (-1)**n
       sumterm = sumterm * factorial((a-b+d+e)/2-n) / factorial(n) / &
            factorial((a-b+c)/2-n)
       sumterm = sumterm * factorial((-b+c+e+f)/2-n) / factorial((d-b+f)/2-n)
       sumterm = sumterm * factorial((a+c+d+f)/2+1-n)/factorial((a+e+f)/2+1-n)
       sumterm = sumterm / factorial((c+d+e)/2+1-n)
       sum=sum+sumterm
    end do

    sixj = phase * sum * outfactors

  end function sixj

  function angdelta(a,b,c)
    use fortrantl
    implicit none
    integer :: a,b,c
    real(dblprec) :: angdelta, scr1
    ! calculate the function delta as defined in varshalovich et al. for
    ! use in 6-j symbol:
    scr1= factorial((a+b-c)/2)
    scr1=scr1/factorial((a+b+c)/2+1)
    scr1=scr1*factorial((a-b+c)/2)
    scr1=scr1*factorial((-a+b+c)/2)
    angdelta=dsqrt(scr1)
  end function angdelta


  function ninej(a,b,c,d,e,f,g,h,i)
    use fortrantl
    implicit none
    integer :: a,b,c,d,e,f,g,h,i
    real(dblprec) :: ninej, sum
    integer :: xlo, xhi
    integer :: x
    ! calculate a 9-j symbol. The arguments are given as integers twice the
    ! value of the true arguments in the form
    ! { a b c }
    ! { d e f }
    ! { g h i }

    ninej=0.0d0
    ! first check for bogus arguments (and return zero if so)
    if(abs(a-b)>c .or. a+b<c) return
    if(abs(d-e)>f .or. d+e<f) return
    if(abs(g-h)>i .or. g+h<i) return
    if(abs(a-d)>g .or. a+d<g) return
    if(abs(b-e)>h .or. b+e<h) return
    if(abs(c-f)>i .or. c+f<i) return

    xlo = max(abs(b-f),abs(a-i),abs(h-d))
    xhi = min(b+f,a+i,h+d)

    sum=0.0d0
    do x=xlo,xhi,2
       sum=sum+(-1)**x*(x+1)*sixj(a,b,c,f,i,x)*sixj(d,e,f,b,x,h)*&
            sixj(g,h,i,x,a,d)
    end do
    ninej=sum

  end function ninej

  recursive function factorial(n) result(res)
    use fortrantl
    implicit none
    integer :: n
    real(dblprec) :: res

    if (n==0 .or. n==1) then
       res=1.0d0
    else
       res=dble(n*factorial(n-1))
    end if
  end function factorial


  recursive function binom(n,r) result(res)
    use fortrantl
    implicit none
    integer :: n,r
    real(dblprec) :: res

    if(n==r .or. r==0) then
       res = 1.0d0
    else if (r==1) then
       res = dble(n)
    else
       res = dble(n)/dble(n-r)*binom(n-1,r)
    end if
  end function binom
end module anglib
