!
! spline --- spline interpolation routines
! Christian Buth
! 28.08.2013
!
! Version 1.0.0: 28.08.2013 (FORTRAN95 module of spline.F and seval.F)
!


      module spline
      use fortrantl
      implicit none


!
! Types
!

      type tspline
        integer n
        real(dblprec), pointer, dimension(:) ::  x, y
        real(dblprec), allocatable ::  b(:), c(:), d(:)
      end type tspline


      contains

!
! initspline --- constructor for a spline object
!
!  the coefficients b(i), c(i), and d(i), i=1,2,...,n are computed
!  for a cubic interpolating spline
!
!    s(x) = y(i) + b(i)*(x-x(i)) + c(i)*(x-x(i))**2 + d(i)*(x-x(i))**3
!
!    for  x(i) .le. x .le. x(i+1)
!
!  input..
!
!    n = the number of data points or knots (n.ge.2)
!    x = the abscissas of the knots in strictly increasing order
!    y = the ordinates of the knots
!
!  output..
!
!    b, c, d  = arrays of spline coefficients as defined above.
!
!  using  p  to denote differentiation,
!
!    y(i) = s(x(i))
!    b(i) = sp(x(i))
!    c(i) = spp(x(i))/2
!    d(i) = sppp(x(i))/6  (derivative from the right)
!
!  the accompanying function subprogram  seval  can be used
!  to evaluate the spline.
!

      subroutine initspline(s, n, x, y)
      implicit none

      integer, intent(in) :: n
      real(dblprec), target, intent(in) :: x(n), y(n)
      type(tspline), intent(out) :: s

      integer nm1, ib, i
      real(dblprec) t

      s%n = n
      s%x => x
      s%y => y
      allocate(s%b(n), s%c(n), s%d(n))

      nm1 = n-1
      if ( n .lt. 2 ) return
      if ( n .lt. 3 ) go to 50
!
!  set up tridiagonal system
!
!  b = diagonal, d = offdiagonal, c = right hand side.
!
      s%d(1) = x(2) - x(1)
      s%c(2) = (y(2) - y(1)) / s%d(1)
      do i = 2, nm1
        s%d(i) = x(i+1) - x(i)
        s%b(i) = 2.0D0*(s%d(i-1) + s%d(i))
        s%c(i+1) = (y(i+1) - y(i)) / s%d(i)
        s%c(i) = s%c(i+1) - s%c(i)
      end do
!
!  end conditions.  third derivatives at  x(1)  and  x(n)
!  obtained from divided differences
!
      s%b(1) = -s%d(1)
      s%b(n) = -s%d(n-1)
      s%c(1) = 0.0D0
      s%c(n) = 0.0D0
      if ( n .eq. 3 ) go to 15
      s%c(1) = s%c(3) / (x(4)-x(2)) - s%c(2) / (x(3)-x(1))
      s%c(n) = s%c(n-1) / (x(n)-x(n-2)) - s%c(n-2) / (x(n-1)-x(n-3))
      s%c(1) = s%c(1)*s%d(1)**2 / (x(4)-x(1))
      s%c(n) = -s%c(n)*s%d(n-1)**2 / (x(n)-x(n-3))
!
!  forward elimination
!
   15 do i = 2, n
        t = s%d(i-1) / s%b(i-1)
        s%b(i) = s%b(i) - t*s%d(i-1)
        s%c(i) = s%c(i) - t*s%c(i-1)
      end do
!
!  back substitution
!
      s%c(n) = s%c(n) / s%b(n)
      do ib = 1, nm1
        i = n-ib
        s%c(i) = (s%c(i) - s%d(i)*s%c(i+1)) / s%b(i)
      end do
!
!  c(i) is now the sigma(i) of the text
!
!  compute polynomial coefficients
!
      s%b(n) = (y(n) - y(nm1)) / s%d(nm1) + s%d(nm1)*(s%c(nm1) + 2.0D0*s%c(n))
      do i = 1, nm1
        s%b(i) = (y(i+1) - y(i)) / s%d(i) - s%d(i)*(s%c(i+1) + 2.0D0*s%c(i))
        s%d(i) = (s%c(i+1) - s%c(i)) / s%d(i)
        s%c(i) = 3.0D0*s%c(i)
      end do
      s%c(n) = 3.0D0*s%c(n)
      s%d(n) = s%d(n-1)
      return

   50 s%b(1) = (y(2)-y(1))/(x(2)-x(1))
      s%c(1) = 0.0D0
      s%d(1) = 0.0D0
      s%b(2) = s%b(1)
      s%c(2) = 0.0D0
      s%d(2) = 0.0D0
      end subroutine initspline


!
! exitspline --- destructor for a spline object
!

      subroutine exitspline(s)
      implicit none

      type(tspline), intent(inout) :: s

      s%n = -1
      s%x => null()
      s%y => null()

      !if (.not. allocated(s%d)) then
      !  stop 'c not allocated!'
      !end if
      !deallocate(s%b, s%c, s%d)
      end subroutine exitspline

!
! spleval --- evaluate a spline object
!
!
!  this function evaluates the cubic spline function
!
!    seval = y(i) + b(i)*(u-x(i)) + c(i)*(u-x(i))**2 + d(i)*(u-x(i))**3
!
!    where  x(i) .lt. u .lt. x(i+1), using horner's rule
!
!  if  u .lt. x(1) then  i = 1  is used.
!  if  u .ge. x(n) then  i = n  is used.
!
!  input..
!
!    n = the number of data points
!    u = the abscissa at which the spline is to be evaluated
!    x,y = the arrays of data abscissas and ordinates
!    b,c,d = arrays of spline coefficients computed by spline
!
!  if  u  is not in the same interval as the previous call, then a
!  binary search is performed to determine the proper interval.
!

      real(dblprec) function spleval(s, u)
      implicit none

      real(dblprec), intent(in) :: u
      type(tspline), intent(in) :: s

      integer i, j, k
      real(dblprec) dx

      data i/1/

      if ( i .ge. s%n ) i = 1
      if ( u .lt. s%x(i) ) go to 10
      if ( u .le. s%x(i+1) ) go to 30
!
!  binary search
!
   10 i = 1
      j = s%n+1
   20 k = (i+j) / 2
      if ( u .lt. s%x(k) ) j = k
      if ( u .ge. s%x(k) ) i = k
      if ( j .gt. i+1 ) go to 20
!
!  evaluate spline
!
   30 dx = u - s%x(i)
      spleval = s%y(i) + dx*(s%b(i) + dx*(s%c(i) + dx*s%d(i)))
      end function spleval
      end module spline
