!
! 2phprop --- propagate an x-ray pulse through a medium
! Christian Buth
! 28 Oct 2009
!
! Version 1.0.0: 01 Nov 2009 (Based on the pulseprop program.  Propagation
!       of x-ray pulse through silicon)
!


!
! Output into FORTRAN units:
! fort.20: initial fundamental x-ray pulse profile
! fort.2?: fundamental x-ray pulse for various steps in the pulse propagation
! fort.30: initial 2nd harmonic x-ray pulse profile
! fort.3?: 2nd harmonic x-ray pulse for various steps in the pulse propagation
! fort.50: initial plot of the number of atoms in the ground state
! fort.97: number of ground-state atoms
! fort.98: number of two-photon absorption events
! fort.99: number of one-photon 2nd harmonic absorption events
! fort.100: sum of fort.98 and fort.99
!


      program twophprop
      use mathtl
      use physconst
      implicit none

      integer, parameter :: ngrid = 10000
      integer i, j

      real(dblprec) Jxnew(ngrid), Jxold(ngrid), Jhnew(ngrid), Jhold(ngrid)
      real(dblprec) z(ngrid), nGS(ngrid), n2ph(ngrid), n2nd(ngrid)
      real(dblprec) nAT, zmin, zmax, L, dz, dt, z0, t0
      real(dblprec) nph, beamarea, tauX, zX, pulseX, strength2nd
      real(dblprec) sigmaXone, sigmaXtwo, sigmaXone2nd, aux, aux1, aux2


      ! Initialization
      write (*,*) '2phprop V1.0.0, two-photon processes in x-ray ',  &
     &            'pulse propagation'
      write (*,*) '(c) 2009 Christian Buth'
      write (*,*)

      write (*,*) 'Number of grid points ', ngrid

      ! Default values for parameters
      ! nAT: number density of neutral atoms in solid, in 1/cm^3. This is
      ! the density before ionization by the x-ray pulse
      nAT = 5.0D22 / CentimeterToBohr**3
      write (*,*) 'Number density ', nAT, 'Bohr^-3 = ',  &
     &            nAT * CentimeterToBohr**3, 'cm^-3'

      ! Length of gas cell
      ! It is assumed that the gas cell starts at z=0 and ends at z=L
      L = 10.0D-4 * CentimeterToBohr
      write (*,*) 'Thickness of foil', L, 'Bohr = ', L  &
     &            * BohrToMicrometer, 'mum'

      ! Smallest z considered.  zmin should be negative
      zmin = -6.0D-4 * CentimeterToBohr
      write (*,*) 'Start grid position', zmin, 'Bohr = ', zmin  &
     &            * BohrToMicrometer, 'mum'

      ! Largest z considered.  zmax should be greater than L
      zmax = 15.0D-4 * CentimeterToBohr
      write (*,*) 'Stop grid postion', zmax, 'Bohr = ', zmax  &
     &            * BohrToMicrometer, 'mum'

      ! Spatial step size
      dz = (zmax - zmin) / real(ngrid-1, dblprec)
      write (*,*) 'Spatial size', zmax - zmin, 'Bohr =', (zmax - zmin)  &
     &            * BohrToMicrometer, 'mum'
      write (*,*) 'Spatial step size', dz, 'Bohr =', dz *  &
     &            BohrToMicrometer, 'mum'

      ! Temporal step size: dt = dz/c
      dt = dz / SpeedOfLight_AU
      write (*,*) 'Temporal step size', dt, 't0 =', dt *  &
     &            t0ToFemtosecond, 'fs'

      ! Initial position of x-ray/laser pulse maximum, in cm.
      ! z0 should be negative, such that both pulses are entirely
      ! outside (to the left of) the gas cell
      z0 = -4.0D-4 * CentimeterToBohr

      ! One-photon x-ray absorption cross section in cm^2 to AU at 2005.8 eV
      !sigmaXone = (1.18D-19 + 6.14D-20 + 7.1561D-21 + 3.9190D-21 + 4.75D-22 + 5.73D-23)  &
      !&            / 1.0D-24 / cross_section_AUtoBarn
      !
      ! At 924 eV: 4.0650*^-20 + 5.0378e-20
      ! At 914 eV: 2.8286*10^-21 + 7.6741*10^-22
      ! Fundamental one-photon x-ray absorption cross section at 914 eV
      sigmaXone = (4.0650D-20 + 5.0378D-20 + 2.8286D-21 + 7.6741D-22)  &
     &            / 1.0D-24 / cross_section_AUtoBarn
      write (*,*) 'Fundamental one-photon cross section ', sigmaXone,  &
     &            'a.u. =', sigmaXone * cross_section_AUtoBarn *       &
     &            1.0D-24, 'cm^2'

      ! Fundamental two-photon x-ray absorption cross section
      sigmaXtwo = 5.0D-56 * CentimeterToBohr**4 * SecondTot0
      write (*,*) 'Fundamental two-photon cross section ', sigmaXtwo,  &
     &            'a.u. =', sigmaXtwo / (CentimeterToBohr**4 *         &
     &            SecondTot0), 'cm^4 s'

      ! 2nd harmonic one-photon x-ray absorption cross section at 1848 eV
      ! 1.8600e+03      1.3762e-19
      ! 1.8161e+03      9.0237e-21
      ! 1.8161e+03      5.4523e-21
      ! 1.8600e+03      5.7121e-22
      ! 1.8600e+03      7.4587e-23
      sigmaXone2nd = (1.3762D-19 + 9.0237D-21 + 5.4523D-21 + 5.7121D-22 +  &
     &                7.4587D-23)  &
     &            / 1.0D-24 / cross_section_AUtoBarn
      write (*,*) '2nd harmonic one-photon cross section ', sigmaXone2nd,  &
     &            'a.u. =', sigmaXone2nd * cross_section_AUtoBarn          &
     &            * 1.0D-24, 'cm^2'

      ! Number of photons in x-ray pulse
      nph = 2.0D12
      write (*,*) 'Number of photons in x-ray pulse', nph

      ! tauX: duration (FWHM) of x-ray pulse, in ps
      tauX = 2.0D0 * FemtosecondTot0
      ! Conversion to length zX = c*tauX
      zX = SpeedOfLight_AU * tauX
      write (*,*) 'Spatial x-ray FWHM size', zX, 'Bohr =', zX *  &
     &            BohrToMicrometer, 'mum'

      ! Construct spatial grid
      z(1) = zmin
      do i = 2, ngrid
        z(i) = z(i-1) + dz
      end do

      ! Initial atomic density and number of two-photon absorption events
      ! along the grid
      do i = 1, ngrid
        if ((z(i) >= 0.0D0) .and. (z(i) <= L)) then
          nGS(i) = nAT
        else
          nGS(i) = 0.0D0
        end if
        n2ph(i) = 0.0D0
        n2nd(i) = 0.0D0
      end do

      ! Write initial atomic density on the grid
      do i = 1, ngrid
        write (50,*) z(i) * BohrToMicrometer, nGS(i)/BohrToCentimeter**3
      end do

      ! Focal beam area
      beamarea = (10.0D0 * MicrometerToBohr)**2
      write (*,*) 'Focal beam area', beamarea, 'a.u. =', beamarea  &
     &            * BohrToMicrometer**2, 'mum^2'

      ! Discretize initial x-ray pulse
      ! X-ray pulse profile (photon flux)
      aux = 2.0D0 * sqrt(log(2.0D0) / M_PI) * nph / tauX / beamarea
      write (*,*) 'Peak x-ray flux', aux, 'a.u. =',               &
     &            aux / (t0ToFemtosecond * BohrToCentimeter**2),  &
     &            'fs^-1 cm^-2'

      ! Strength of 2nd harmonic in LCLS beam
      strength2nd = 1.0D-2
      write (*,*) 'Strength of the 2nd harmonic radiation ',  &
     &            '(fraction of the fundamental) ', strength2nd

      ! Discretize and write out initial x-ray pulse
      do i = 1, ngrid
        Jxold(i) = pulseX(z(i), z0, zX, aux)
        write (20,*) z(i) * BohrToMicrometer, Jxold(i)  &
     &               / (t0ToFemtosecond * BohrToCentimeter**2)
        Jhold(i) = Jxold(i) * strength2nd
        write (30,*) z(i) * BohrToMicrometer, Jhold(i)  &
     &               / (t0ToFemtosecond * BohrToCentimeter**2)
      end do

      t0 = 0.0D0
      ! Propagation of the pulses over the grid
      do j = 1, ngrid
        t0 = t0 + dt
        z0 = z0 + dz

        ! Reduction of the x-ray flux due to absorption
        Jxnew(1) = 0.0D0                ! Fill the left-most cell with zero
        Jhnew(1) = 0.0D0
        do i = 2, ngrid
          aux1 = nGS(i-1) * sigmaXone * dz
          aux2 = nGS(i-1) * sigmaXtwo * Jxold(i-1) * dz
          aux = 1.0D0 - aux1 - aux2
          Jxnew(i) = aux * Jxold(i-1)
          aux1 = nGS(i-1) * sigmaXone2nd * dz
          aux = 1.0D0 - aux1
          Jhnew(i) = aux * Jhold(i-1)
        end do

        ! Reduction of the ground-state atom number density
        do i = 1, ngrid
          if ((z(i) >= 0.0D0) .and. (z(i) <= L)) then
            aux = nGS(i) * sigmaXtwo * Jxold(i)**2 * dt
            n2ph(i) = n2ph(i) + beamarea * dz * aux
            aux = aux + nGS(i) * sigmaXone * Jxold(i) * dt
            aux1 = nGS(i) * sigmaXone2nd * Jhold(i) * dt
            n2nd(i) = n2nd(i) + beamarea * dz * aux1
            aux = aux + aux1
            nGS (i) = nGS (i) - aux
            if (nGS (i) < 0.0D0) then
              nGS (i) = 0.0D0
            end if
          end if
        end do

        ! Write snapshots of the pulse propagation
        ! Before the foil
        if (abs(z0 - (-3.0D-4) * CentimeterToBohr) < dz / 2.0D0) then
          write (*,*) 'Grid position -3mum, ie.,', j, 'time steps'
          do i = 1, ngrid
             if (mod(i,10) == 0) then
               write (21,*) z(i) * BohrToMicrometer, Jxnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
               write (31,*) z(i) * BohrToMicrometer, Jhnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
             end if
          end do
        end if
        ! In the foil
        if (abs(z0 - 5.0D-4 * CentimeterToBohr) < dz / 2.0D0) then
          write (*,*) 'Grid position 5mum, ie.,', j, 'time steps'
          do i = 1, ngrid
             if (mod(i,10) == 0) then
               write (22,*) z(i) * BohrToMicrometer, Jxnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
               write (32,*) z(i) * BohrToMicrometer, Jhnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
             end if
          end do
        end if
        ! After the foil
        if (abs(z0 - 14.0D-4 * CentimeterToBohr) < dz / 2.0D0) then
          write (*,*) 'Grid position 14mum, ie.,', j, 'time steps'
          do i = 1, ngrid
             if (mod(i,10) == 0) then
               write (23,*) z(i) * BohrToMicrometer, Jxnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
               write (33,*) z(i) * BohrToMicrometer, Jhnew(i)  &
     &                      / (t0ToFemtosecond * BohrToCentimeter**2)
             end if
          end do
        end if

        Jxold(:) = Jxnew(:)
        Jhold(:) = Jhnew(:)
      end do

      ! Output grid after pulse has gone through the gas cell
      do i = 1, ngrid
        write (97,*) z(i) * BohrToMicrometer, nGS(i)/BohrToCentimeter**3
        write (98,*) z(i) * BohrToMicrometer, n2ph(i)
        write (99,*) z(i) * BohrToMicrometer, n2nd(i)
        write (100,*) z(i) * BohrToMicrometer, n2ph(i) + n2nd(i)
      end do

      write (*,*)
      write (*,*) 'All pulses propagated successfully!'
      end program twophprop


!
! pulseX --- the x-ray pulse
!

      function pulseX(z, z0, zL, J0)
      use fortrantl
      implicit none

      real(dblprec) pulseX, z, z0, zL, J0
      real(dblprec), parameter :: aux = 4.0D0 * log(2.0D0)

      pulseX = J0 * exp(-aux*((z-z0)/zL)**2)
      end function pulseX
